// -------------------------------------------------------------------------------- 
//!  @file  bitmap.c                                                                
//!  ビットマップ処理のソースファイル                                               
//                                                                                  
//   Copyright (C) 2024 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "bitmap.h"


// -------------------------------------------------------------------------------- 
//!  ビットマップファイルをロード                                                   
//!                                                                                 
//!  @param [in]  p_fileName  ファイル名                                            
//!  @retval      NULL以外    ビットマップ情報                                      
//!  @retval      NULL        異常終了                                              
// -------------------------------------------------------------------------------- 
BITMAPINFO *bitmap_load( const char *p_fileName )
{
	struct stat			fileStat ;
	BITMAPFILEHEADER	bfh ;
	BITMAPINFOHEADER	bih ;
	BITMAPINFO			*p_bi ;
	uint8_t				*p_data ;
	int					result, fd, index, size, line, pad, width, height ;


	fd = open(&p_fileName[0], O_RDONLY) ;
	if( fd < 0 ) {
		fprintf(stderr, "******** %s(%d) : can't open %s\n", __func__, __LINE__, &p_fileName[0]) ;
		return NULL ;
	}

	result = -1 ;
	p_bi   = NULL ;

	do {
		// ファイルサイズ取得
		if( fstat(fd, &fileStat) < 0 ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		// ビットマップファイルヘッダ
		if( read(fd, &bfh, sizeof( bfh )) != sizeof( bfh ) ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		if( bfh.bfType != 0x4d42 ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		if( bfh.bfSize > fileStat.st_size ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		if( (bfh.bfOffBits<sizeof( bfh )) || (bfh.bfOffBits>=fileStat.st_size) ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		// ビットマップ情報ヘッダ
		if( read(fd, &bih, sizeof( bih )) != sizeof( bih ) ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		if( bih.biSize < sizeof( BITMAPFILEHEADER ) ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		if( bih.biSize != sizeof( BITMAPFILEHEADER ) ) {
			lseek(fd, bih.biSize, SEEK_CUR) ;
		}

		width  = bih.biWidth ;
		height = bih.biHeight ;
		if( height < 0 ) {
			bih.biHeight = -bih.biHeight ;
			width        = -bih.biWidth ;
		}

		if( (bih.biWidth<=0) || !bih.biHeight ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		if( bih.biPlanes != 1 ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		if( bih.biBitCount != 8 ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		if( bih.biCompression != BI_RGB ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		line = (bih.biWidth + 3) & ~3 ;
		pad  = line - bih.biWidth ;

		if( bih.biSizeImage ) {
			if( bih.biSizeImage != (line*bih.biHeight) ) {
				fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
				break ;
			}
		}
		bih.biSizeImage = bih.biWidth * bih.biHeight ;

		if( bih.biClrUsed > 256 ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		if( !bih.biClrUsed ) {
			bih.biClrUsed = 256 ;
		}

		if( bfh.bfOffBits < (bih.biSize+sizeof( BITMAPFILEHEADER )+(bih.biClrUsed*sizeof( RGBQUAD ))) ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		// メモリ確保
		size = sizeof( BITMAPINFOHEADER ) + (bih.biClrUsed * sizeof( RGBQUAD )) + bih.biSizeImage ;
		p_bi = (BITMAPINFO *)malloc( size ) ;
		if( p_bi == NULL ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}
		memset(p_bi, 0, size) ;
		memcpy(&p_bi->bmiHeader, &bih, sizeof( p_bi->bmiHeader )) ;
		p_bi->bmiHeader.biSize = sizeof( BITMAPFILEHEADER ) ;

		// カラーテーブル
		lseek(fd, sizeof( BITMAPFILEHEADER )+bih.biSize, SEEK_SET) ;

		if( read(fd, &p_bi->bmiColors[0], bih.biClrUsed*sizeof( RGBQUAD )) != (bih.biClrUsed*sizeof( RGBQUAD )) ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		for(index=0; index<bih.biClrUsed; index++) {
			p_bi->bmiColors[index].rgbReserved = 0 ;
		}

		// 画像データ
		lseek(fd, bfh.bfOffBits, SEEK_SET) ;

		p_data = (uint8_t *)&p_bi->bmiColors[bih.biClrUsed] ;
		if( height > 0 ) {
			p_data = &p_data[bih.biWidth*(bih.biHeight-1)] ;
		}

		for(index=0; index<bih.biHeight; index++) {
			if( read(fd, &p_data[0], bih.biWidth) != bih.biWidth ) {
				fprintf(stderr, "******** %s(%d) : %d\n", __func__, __LINE__, index) ;
				break ;
			}
			p_data -= width ;

			if( pad ) {
				lseek(fd, pad, SEEK_CUR) ;
			}
		}
		if( index != bih.biHeight ) {
			break ;
		}

		// 正常終了
		result = 0 ;
	} while( 0 ) ;

	close( fd ) ;

	if( result ) {
		if( p_bi != NULL ) {
			free( p_bi ) ;
			p_bi = NULL ;
		}
	}

	return p_bi ;
}


// -------------------------------------------------------------------------------- 
//!  ビットマップ情報を解放                                                         
//!                                                                                 
//!  @param [in,out] p_bi  ビットマップ情報                                         
//!  @retval          0    正常終了                                                 
//!  @retval         -1    異常終了                                                 
// -------------------------------------------------------------------------------- 
int bitmap_free( BITMAPINFO *p_bi )
{
	if( p_bi == NULL ) {
		return -1 ;
	}

	free( p_bi ) ;

	return 0 ;
}
