// -------------------------------------------------------------------------------- 
//!  @file  common.h                                                                
//!  共通処理のソースファイル                                                       
//                                                                                  
//   Copyright (C) 2024 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <unistd.h>
#include <libgen.h>
#include "common.h"
#include "inifile.h"


SETUPINFO							g_setupInfo ;					//!< 設定情報


// -------------------------------------------------------------------------------- 
//!  パラメータを取得                                                               
//!                                                                                 
//!  @param [in]  p_string  文字列                                                  
//!  @param [out] pp_param  パラメータアドレスの格納アドレス                        
//!  @retval      NULL      異常終了                                                
//!  @retval      上記以外  パラメータ取得後の文字列                                
// -------------------------------------------------------------------------------- 
static char *getParam(char *p_string, char **pp_param)
{
	if( p_string == NULL ) {
		*pp_param = NULL ;
		return NULL ;
	}

	p_string = ltrim( &p_string[0] ) ;

	*pp_param = &p_string[0] ;

	p_string = strchr(&p_string[0], ',') ;
	if( p_string == NULL ) {
		*pp_param = NULL ;
		return NULL ;
	}

	*p_string++ = '\0' ;

	return &p_string[0] ;
}


// -------------------------------------------------------------------------------- 
//!  ラジオ局情報を作成                                                             
//!                                                                                 
//!  @param [in]  p_sectionInfo   セクション情報                                    
//!  @retval       0              正常終了                                          
//!  @retval      -1              異常終了                                          
// -------------------------------------------------------------------------------- 
static int createStationInfo( INISECTIONINFO *p_sectionInfo )
{
	STATIONINFO		*p_stationInfo ;
	INIKEYINFO		*p_keyInfo ;
	char			*p_buff, *p_id, *p_image, *p_end ;
	int				index ;

	// ラジオ局なし
	if( !p_sectionInfo->keyInfoCount ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	// ラジオ情報確保
	p_stationInfo = malloc( p_sectionInfo->keyInfoCount * sizeof( STATIONINFO ) ) ;
	if( p_stationInfo == NULL ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	p_keyInfo = &p_sectionInfo->p_keyInfo[0] ;

	for(index=0; index<p_sectionInfo->keyInfoCount; index++) {
		p_buff = getParam(&p_keyInfo->p_param[0], &p_id) ;
		p_buff = getParam(&p_buff[0],             &p_image) ;

		if( (p_id==NULL) || (p_image==NULL) ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		strcpy(&p_stationInfo[index].type[0], &p_keyInfo->p_name[0]) ;
		strcpy(&p_stationInfo[index].id[0],   &p_id[0]) ;
		p_stationInfo[index].image = strtoul(&p_image[0], &p_end, 0) ;
		p_keyInfo++ ;
	}

	if( index < p_sectionInfo->keyInfoCount ) {
		free( p_stationInfo ) ;
		return -1 ;
	}

	g_setupInfo.p_stationInfo    = &p_stationInfo[0] ;
	g_setupInfo.stationInfoCount = p_sectionInfo->keyInfoCount ;

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  行頭の空白を削除                                                               
//!                                                                                 
//!  @param [in] p_string  文字列                                                   
//!  @return               削除後の文字列                                           
// -------------------------------------------------------------------------------- 
char *ltrim( char *p_string )
{
	while( isspace( *p_string ) ) {
		p_string++ ;
	}

	return &p_string[0] ;
}


// -------------------------------------------------------------------------------- 
//!  行末の空白を削除                                                               
//!                                                                                 
//!  @param [in] p_string  文字列アドレス                                           
//!  @return               なし                                                     
// -------------------------------------------------------------------------------- 
void rtrim( char *p_string )
{
	char	*p_last ;


	p_last = "\0" ;

	while( *p_string != '\0' ) {
		if( !isspace( *p_string ) ) {
			p_last = &p_string[1] ;
		}
		p_string++ ;
	}

	if( *p_last != '\0' ) {
		*p_last = '\0' ;
	}
}


// -------------------------------------------------------------------------------- 
//!  行頭・行末の空白を削除                                                         
//!                                                                                 
//!  @param [in] p_string  文字列                                                   
//!  @return               削除後の文字列                                           
// -------------------------------------------------------------------------------- 
char *trim( char *p_string )
{
	p_string = ltrim( &p_string[0] ) ;
	rtrim( &p_string[0] ) ;

	return &p_string[0] ;
}


// -------------------------------------------------------------------------------- 
//!  ミリ秒待ち                                                                     
//!                                                                                 
//!  @param [in] msec  ミリ秒                                                       
//!  @return           なし                                                         
// -------------------------------------------------------------------------------- 
void msleep( int msec )
{
	struct timespec		ts ;


	ts.tv_sec  =  msec / 1000 ;
	ts.tv_nsec = (msec % 1000) * 1000 * 1000 ;
	nanosleep(&ts, NULL) ;
}


// -------------------------------------------------------------------------------- 
//!  カレント時間を取得                                                             
//!                                                                                 
//!  @return ミリ秒                                                                 
// -------------------------------------------------------------------------------- 
uint32_t getCurrentTime( void )
{
	struct timespec		ts ;


	clock_gettime(CLOCK_MONOTONIC_RAW, &ts) ;

	return (uint32_t )((ts.tv_sec * 1000) + (ts.tv_nsec / (1000 * 1000))) ;
}


// -------------------------------------------------------------------------------- 
//!  パスからカレントディレクトを設定                                               
//!                                                                                 
//!  @param [in] p_path  パス                                                       
//!  @return             なし                                                       
// -------------------------------------------------------------------------------- 
void setCurrentDir( const char *p_path )
{
	char	path[PATH_MAX] ;


	strcpy(&path[0], &p_path[0]) ;

	chdir( dirname( &path[0] ) ) ;
}


// -------------------------------------------------------------------------------- 
//!  設定ファイルをロード                                                           
//!                                                                                 
//!  @param [in] p_fileName  設定ファイル名                                         
//!  @retval      0          正常終了                                               
//!  @retval     -1          異常終了                                               
// -------------------------------------------------------------------------------- 
int loadSetupFile( const char *p_fileName )
{
	INIFILEINFO		*p_fileInfo ;
	INISECTIONINFO	*p_sectionInfo ;
	int				result, value ;


	memset(&g_setupInfo, 0, sizeof( g_setupInfo )) ;

	p_fileInfo = initfile_open( &p_fileName[0] ) ;
	if( p_fileInfo == NULL ) {
		return -1 ;
	}

	result = -1 ;

	do {
		// 再生スクリプト
		if( initfile_getString(p_fileInfo, "setup", "script_play", &g_setupInfo.script.play[0], sizeof( g_setupInfo.script.play )) == false ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		// 録音スクリプト
		if( initfile_getString(p_fileInfo, "setup", "script_rec", &g_setupInfo.script.record[0], sizeof( g_setupInfo.script.record )) == false ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		// 録音ファイル出力ディレクトリ
		if( initfile_getString(p_fileInfo, "setup", "output_rec", &g_setupInfo.outputRecord[0], sizeof( g_setupInfo.outputRecord )) == false ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		// I2C デバイスアドレス
		if( initfile_getString(p_fileInfo, "setup", "i2c_dev", &g_setupInfo.i2cDev[0], sizeof( g_setupInfo.i2cDev )) == false ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		// ピン
		if( initfile_getValue(p_fileInfo, "setup", "pin_green", 1, 40, &g_setupInfo.pin[PIN_GREEN]) == false ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		if( initfile_getValue(p_fileInfo, "setup", "pin_blue", 1, 40, &g_setupInfo.pin[PIN_BLUE]) == false ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		if( initfile_getValue(p_fileInfo, "setup", "pin_yellow", 1, 40, &g_setupInfo.pin[PIN_YELLOW]) == false ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		if( initfile_getValue(p_fileInfo, "setup", "pin_red", 1, 40, &g_setupInfo.pin[PIN_RED]) == false ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		// Bluetooth デバイスアドレス
		if( initfile_getString(p_fileInfo, "setup", "bdaddr", &g_setupInfo.btAddr[0], sizeof( g_setupInfo.btAddr )) == false ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		// WLAN インタフェース
		if( initfile_getString(p_fileInfo, "setup", "wlanif", &g_setupInfo.wlanIf[0], sizeof( g_setupInfo.wlanIf )) == false ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		// ミキサー
		if( initfile_getString(p_fileInfo, "setup", "mixer", &g_setupInfo.mixer[0], sizeof( g_setupInfo.mixer )) == false ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		// シンプルミキサー要素
		if( initfile_getString(p_fileInfo, "setup", "selem", &g_setupInfo.selem[0], sizeof( g_setupInfo.selem )) == false ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		// コントラスト
		if( initfile_getValue(p_fileInfo, "setup", "contrast", 1, 255, &g_setupInfo.contrast) == false ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		// 表示オフカウンタ
		if( initfile_getValue(p_fileInfo, "setup", "displayoff", 3, 24*60*60, &value) == false ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}
		g_setupInfo.displayOffCount = value ;

		// ラジオ局情報
		p_sectionInfo = initfile_getSection(p_fileInfo, "station") ;
		if( p_sectionInfo == NULL ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		result = createStationInfo( p_sectionInfo ) ;
		if( result < 0 ) {
			break ;
		}

		// 正常終了
		result = 0 ;
	} while( 0 ) ;

	initfile_close( p_fileInfo ) ;

	return result ;
}
