// -------------------------------------------------------------------------------- 
//!  @file  gpio.c                                                                  
//!  GPIO 処理のソースファイル                                                      
//                                                                                  
//   Copyright (C) 2024 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <gpiod.h>
#include "common.h"


static struct gpiod_chip			*m_p_chip=NULL ;				//!< チップハンドル
static struct gpiod_line			*m_p_line[PIN_MAX] ;			//!< ラインハンドル


// -------------------------------------------------------------------------------- 
//!  開く                                                                           
//!                                                                                 
//!  @retval  0  正常終了                                                           
//!  @retval -1  異常終了                                                           
// -------------------------------------------------------------------------------- 
int gpio_open( void )
{
	struct gpiod_chip	*p_chip ;
	char				buff[64] ;
	int					result, index ;


	if( m_p_chip != NULL ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	p_chip = gpiod_chip_open_lookup( "" ) ;
	if( p_chip == NULL ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	for(index=0; index<PIN_MAX; index++) {
		sprintf(&buff[0], "raspi-gpio set %d ip pu", g_setupInfo.pin[index]) ;
		result = system( &buff[0] ) ;
		if( result ) {
			fprintf(stderr, "******** %s(%d) : %d\n", __func__, __LINE__, result) ;
		}

		m_p_line[index] = gpiod_chip_get_line(p_chip, g_setupInfo.pin[index]) ;
		if( m_p_line[index] == NULL ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		if( gpiod_line_request_input(m_p_line[index], "xxxx") != 0 ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}
	}
	if( index < PIN_RED ) {
		gpiod_chip_close( p_chip ) ;
		return -1 ;
	}

	m_p_chip = p_chip ;

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  閉じる                                                                         
//!                                                                                 
//!  @retval  0  正常終了                                                           
//!  @retval -1  異常終了                                                           
// -------------------------------------------------------------------------------- 
int gpio_close( void )
{
	if( m_p_chip == NULL ) {
		return -1 ;
	}

	gpiod_chip_close( m_p_chip ) ;

	m_p_chip = NULL ;

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  ボタン状態を取得                                                               
//!                                                                                 
//!  @return ボタン状態                                                             
// -------------------------------------------------------------------------------- 
uint32_t gpio_button( void )
{
	uint32_t	status ;
	int			index, value ;


	if( m_p_chip == NULL ) {
		return 0 ;
	}

	status = 0 ;

	for(index=0; index<countof( m_p_line ); index++) {
		value = gpiod_line_get_value( m_p_line[index] ) ;
		if( !value ) {
			status |= (1 << index) ;
		}
	}

	return status ;
}
