// -------------------------------------------------------------------------------- 
//!  @file  inifile.c                                                               
//!  ini ファイル制御のソースファイル                                               
//                                                                                  
//   Copyright (C) 2024 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include "common.h"
#include "inifile.h"


// -------------------------------------------------------------------------------- 
//!  バッファから１行の文字列を取得                                                 
//!                                                                                 
//!  @param [in,out] pp_buff  作業バッファ格納アドレス                              
//!  @return                  文字列アドレス                                        
// -------------------------------------------------------------------------------- 
static char *getLine( char **pp_buff )
{
	char	*p_buff, *p_string ;


	p_string = *pp_buff ;
	p_buff   = &p_string[0] ;

	while( 1 ) {
		// 終端
		if( *p_buff == '\0' ) {
			break ;
		}

		// 改行
		if( *p_buff == '\n' ) {
			*p_buff++ = '\0' ;

			if( *p_buff == '\r' ) {
				p_buff++ ;
			}
			break ;
		}

		if( *p_buff == '\r' ) {
			*p_buff++ = '\0' ;
			if( *p_buff == '\n' ) {
				p_buff++ ;
			}
			break ;
		}

		// コメント
		if( (*p_buff==';') || (*p_buff=='#') ) {
			*p_buff = '\0' ;
		}

		p_buff++ ;
	}

	*pp_buff = &p_buff[0] ;

	return &p_string[0] ;
}


// -------------------------------------------------------------------------------- 
//!  セクション文字列を解析                                                         
//!                                                                                 
//!  @param [in,out] p_fileInfo  ini ファイル情報                                   
//!  @param [in]     line        行                                                 
//!  @param [in,out] p_string    文字列                                             
//!  @retval          0          正常終了                                           
//!  @retval         -1          異常終了                                           
// -------------------------------------------------------------------------------- 
static int parseSection(INIFILEINFO *p_fileInfo, int line, char *p_string)
{
	INISECTIONINFO	*p_sectionInfo ;


	if( p_fileInfo->sectionInfoCount == INIFILE_SECTION_MAX ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	p_sectionInfo               = &p_fileInfo->sectionInfo[p_fileInfo->sectionInfoCount++] ;
	p_sectionInfo->line         = line ;
	p_sectionInfo->p_keyInfo    = &p_fileInfo->keyInfo[p_fileInfo->keyInfoCount] ;
	p_sectionInfo->keyInfoCount = 0 ;

	// 名前
	p_sectionInfo->p_name = &p_string[1] ;

	p_string = strchr(&p_string[1], ']') ;
	if( p_string == NULL ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}
	*p_string++ = '\0' ;

	if( *p_sectionInfo->p_name == '\0' ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	if( *p_string != '\0' ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  キー文字列を解析                                                               
//!                                                                                 
//!  @param [in,out] p_fileInfo  ini ファイル情報                                   
//!  @param [in]     line        行                                                 
//!  @param [in,out] p_string    文字列                                             
//!  @retval          0          正常終了                                           
//!  @retval         -1          異常終了                                           
// -------------------------------------------------------------------------------- 
static int parseKey(INIFILEINFO *p_fileInfo, int line, char *p_string)
{
	INIKEYINFO	*p_keyInfo ;


	if( !p_fileInfo->sectionInfoCount ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	if( p_fileInfo->keyInfoCount == INIFILE_KEY_MAX ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	p_keyInfo                = &p_fileInfo->keyInfo[p_fileInfo->keyInfoCount++] ;
	p_keyInfo->p_sectionInfo = &p_fileInfo->sectionInfo[p_fileInfo->sectionInfoCount-1] ;
	p_keyInfo->line          = line ;

	p_keyInfo->p_sectionInfo->keyInfoCount++ ;

	// 名前
	p_keyInfo->p_name = &p_string[0] ;

	p_string = strchr(&p_string[0], '=') ;
	if( p_string == NULL ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}
	*p_string++ = '\0' ;

	rtrim( (char *)&p_keyInfo->p_name[0] ) ;

	if( p_keyInfo->p_name[0] == '\0' ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	// パラメータ
	p_keyInfo->p_param = ltrim( &p_string[0] ) ;

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  バッファを解析                                                                 
//!                                                                                 
//!  @param [in,out] p_fileInfo  ini ファイル情報                                   
//!  @param [in]     p_buff      作業バッファ                                       
//!  @retval          0          正常終了                                           
//!  @retval         -1          異常終了                                           
// -------------------------------------------------------------------------------- 
static int parseText(INIFILEINFO *p_fileInfo, char *p_buff)
{
	char	*p_line ;
	int		result, line ;


	line = 0 ;

	while( *p_buff != '\0' ) {
		line++ ;

		p_line = getLine( &p_buff ) ;
		trim( &p_line[0] ) ;

		// 空白行
		if( p_line[0] == '\0' ) {
			continue ;
		}

		// コメント
		if( (p_line[0]==';') || (p_line[0]=='#') ) {
			continue ;
		}

		if( p_line[0] == '[' ) {	// セクション
			result = parseSection(p_fileInfo, line, &p_line[0]) ;
		} else {					// キー
			result = parseKey(p_fileInfo, line, &p_line[0]) ;
		}
		if( result < 0 ) {
			break ;
		}
	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  開く                                                                           
//!                                                                                 
//!  @param [in] p_fileName  ファイル名                                             
//!  @retval     NULL        異常終了                                               
//!  @retval     上記以外    ini ファイル情報                                       
// -------------------------------------------------------------------------------- 
INIFILEINFO *initfile_open( const char *p_fileName )
{
	INIFILEINFO		*p_fileInfo ;
	struct stat		stat ;
	int				result, fd, size ;


	fd = open(&p_fileName[0], O_RDWR) ;
	if( fd == -1 ) {
		fprintf(stderr, "******** %s(%d) : can't open %s\n", __func__, __LINE__, &p_fileName[0]) ;
		return NULL ;
	}

	p_fileInfo = NULL ;
	result     = -1 ;

	do {
		if( fstat(fd, &stat) < 0 ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		if( !stat.st_size ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		size       = sizeof( INIFILEINFO ) + stat.st_size + 1 ;
		p_fileInfo = (INIFILEINFO *)malloc( size ) ;
		if( p_fileInfo == NULL ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}
		memset(p_fileInfo, 0, size) ;

		result = read(fd, &p_fileInfo[1], stat.st_size) ;
		if( result != stat.st_size ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			result = -1 ;
			break ;
		}

		result = parseText(p_fileInfo, (char *)&p_fileInfo[1]) ;
	} while( 0 ) ;

	close( fd ) ;

	if( result ) {
		if( p_fileInfo != NULL ) {
			free( p_fileInfo ) ;
			p_fileInfo = NULL ;
		}
	}

	return p_fileInfo ;
}


// -------------------------------------------------------------------------------- 
//!  閉じる                                                                         
//!                                                                                 
//!  @param [in] p_fileInfo  ini ファイル情報                                       
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
void initfile_close( INIFILEINFO *p_fileInfo )
{
	free( p_fileInfo ) ;
}


// -------------------------------------------------------------------------------- 
//!  セクション情報を取得                                                           
//!                                                                                 
//!  @param [in] p_fileInfo  ini ファイル情報                                       
//!  @param [in] p_name      セクション名                                           
//!  @retval     NULL        セクション情報なし                                     
//!  @retval     上記以外    セクション情報                                         
// -------------------------------------------------------------------------------- 
INISECTIONINFO *initfile_getSection(INIFILEINFO *p_fileInfo, const char *p_name)
{
	INISECTIONINFO	*p_sectionInfo, *p_sectionInfoEnd ;


	p_sectionInfo    = &p_fileInfo->sectionInfo[0] ;
	p_sectionInfoEnd = &p_fileInfo->sectionInfo[p_fileInfo->sectionInfoCount] ;

	while( p_sectionInfo != p_sectionInfoEnd ) {
		if( !strcmp(&p_sectionInfo->p_name[0], &p_name[0]) ) {
			return p_sectionInfo ;
		}
		p_sectionInfo++  ;
	}

	return NULL ;
}


// -------------------------------------------------------------------------------- 
//!  キー情報を取得                                                                 
//!                                                                                 
//!  @param [in] p_sectionInfo  セクション情報                                      
//!  @param [in] p_keyInfo         NULL:先頭から検索                                
//!                             上記以外:前回のキー情報                             
//!  @param [in] p_name         キー名                                              
//!  @retval     NULL           キー情報なし                                        
//!  @retval     上記以外       キー情報                                            
// -------------------------------------------------------------------------------- 
INIKEYINFO *initfile_getKey(INISECTIONINFO *p_sectionInfo, INIKEYINFO *p_keyInfo, const char *p_name)
{
	INIKEYINFO	*p_keyInfoEnd ;


	if( p_keyInfo == NULL ) {
		p_keyInfo = &p_sectionInfo->p_keyInfo[0] ;
	} else {
		p_keyInfo++ ;
	}

	p_keyInfoEnd = &p_sectionInfo->p_keyInfo[p_sectionInfo->keyInfoCount] ;

	while( p_keyInfo < p_keyInfoEnd ) {
		if( !strcmp(&p_keyInfo->p_name[0], &p_name[0]) ) {
			return p_keyInfo ;
		}
		p_keyInfo++ ;
	}

	return NULL ;
}


// -------------------------------------------------------------------------------- 
//!  セクション・キー名からキー情報を取得                                           
//!                                                                                 
//!  @param [in] p_fileInfo     ini ファイル情報                                    
//!  @param [in] p_sectionName  セクション名                                        
//!  @param [in] p_keyName      キー名                                              
//!  @retval     NULL           キー情報なし                                        
//!  @retval     上記以外       キー情報                                            
// -------------------------------------------------------------------------------- 
INIKEYINFO *initfile_getSectionKey(INIFILEINFO *p_fileInfo, const char *p_sectionName, const char *p_keyName)
{
	INISECTIONINFO	*p_sectionInfo ;


	p_sectionInfo = initfile_getSection(p_fileInfo, &p_sectionName[0]) ;
	if( p_sectionInfo == NULL ) {
		return NULL ;
	}

	return initfile_getKey(p_sectionInfo, NULL, &p_keyName[0]) ;
}


// -------------------------------------------------------------------------------- 
//!  文字列を取得                                                                   
//!                                                                                 
//!  @param [in]  p_fileInfo     ini ファイル情報                                   
//!  @param [in]  p_sectionName  セクション名                                       
//!  @param [in]  p_keyName      キー名                                             
//!  @param [out] p_string       文字列格納アドレス                                 
//!  @param [in]  size           文字列格納サイズ                                   
//!  @retval      true           正常終了                                           
//!  @retval      false          異常終了                                           
// -------------------------------------------------------------------------------- 
bool initfile_getString(INIFILEINFO *p_fileInfo, const char *p_sectionName, const char *p_keyName, char *p_string, int size)
{
	INIKEYINFO	*p_keyInfo ;
	size_t		length ;


	p_keyInfo = initfile_getSectionKey(p_fileInfo, &p_sectionName[0], &p_keyName[0]) ;
	if( p_keyInfo == NULL ) {
		return false ;
	}

	length = strlen( &p_keyInfo->p_param[0] ) ;
	if( length > (size+1) ) {
		return false ;
	}

	strcpy(&p_string[0], &p_keyInfo->p_param[0]) ;

	return true ;
}


// -------------------------------------------------------------------------------- 
//!  値を取得                                                                       
//!                                                                                 
//!  @param [in]  p_fileInfo     ini ファイル情報                                   
//!  @param [in]  p_sectionName  セクション名                                       
//!  @param [in]  p_keyName      キー名                                             
//!  @param [in]  min            最小値                                             
//!  @param [in]  max            最大値                                             
//!  @param [out] p_value        値格納アドレス                                     
//!  @retval      true           正常終了                                           
//!  @retval      false          異常終了                                           
// -------------------------------------------------------------------------------- 
bool initfile_getValue(INIFILEINFO *p_fileInfo, const char *p_sectionName, const char *p_keyName, int min, int max, int *p_value)
{
	INIKEYINFO	*p_keyInfo ;
	char		*p_end ;
	long		value ;


	p_keyInfo = initfile_getSectionKey(p_fileInfo, &p_sectionName[0], &p_keyName[0]) ;
	if( p_keyInfo == NULL ) {
		return false ;
	}

	value = strtol(&p_keyInfo->p_param[0], &p_end, 0) ;
	if( *p_end != '\0' ) {
		return false ;
	}

	if( (value<min) || (value>max) ) {
		return false ;
	}

	*p_value = (int )value ;

	return true ;
}
