// -------------------------------------------------------------------------------- 
//!  @file  main.c                                                                  
//!  メイン処理のソースファイル                                                     
//                                                                                  
//   Copyright (C) 2024 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <locale.h>
#include <time.h>
#include <libgen.h>
#include <signal.h>
#include <termios.h>
#include <unistd.h>
#include "common.h"
#include "bitmap.h"
#include "inifile.h"
#include "gpio.h"
#include "bluetooth.h"
#include "wlan.h"
#include "sound.h"
#include "ssd1306.h"

#define CPUTEMP_NAME				"/sys/class/thermal/thermal_zone0/temp"			//!< CPU 温度ファイル名
#define MSECTOCOUNT( msec )			(((uint32_t )(msec)+(PERIOD_MS/2))/PERIOD_MS)	//!< ミリ秒からカウント数変換


typedef uint32_t (*TIMERPROC)( void ) ;								//!< タイマー処理
typedef void     (*MAINPROC)( void ) ;								//!< メイン処理
typedef void     (*MENUITEMPROC)( void ) ;							//!< メニュー項目選択処理

typedef struct _menuinfo			MENUINFO ;						//!< メニュー情報


	enum {
		FIXP_ONE					=	256,						//!< 固定小数の１
		PERIOD_MS					=	30,							//!< メイン処理の周期ミリ秒
		DISPLAYOFF_COUNT			=	MSECTOCOUNT(90000),			//!< 表示オフカウンタ
		DISPLAYOFF_WAIT				=	100,						//!< 表示オフ時の待ちミリ秒
		CONTRAST_COUNT				=	DISPLAYOFF_COUNT - MSECTOCOUNT(30000),
																	//!< 無操作時のコントラスト変更カウンタ
		BUTTON_REPEAT				=	MSECTOCOUNT(500),			//!< ボタン繰り返し待ち回数
		PHASE_DATETIME				=	MSECTOCOUNT(1000),			//!< 日時タイマー起動位相
		PHASE_BLUETOOTH				=	MSECTOCOUNT(60),			//!< Bluetooth 接続状態タイマー起動位相
		PHASE_WLAN					=	MSECTOCOUNT(90),			//!< LAN 接続状態タイマー起動位相
		PHASE_VOLUME				=	MSECTOCOUNT(120),			//!< 音量タイマー起動位相
		PHASE_PLAY					=	MSECTOCOUNT(150),			//!< 再生状態タイマー起動位相
		PHASE_RECORD				=	MSECTOCOUNT(180),			//!< 録音状態タイマー起動位相
		PHASE_CPUTEMP				=	MSECTOCOUNT(210),			//!< CPU 温度起動位相
		PERIOD_DATETIME				=	MSECTOCOUNT(750),			//!< 日時タイマー起動周期
		PERIOD_BLUETOOTH			=	MSECTOCOUNT(400),			//!< Bluetooth 接続状態タイマー起動周期
		PERIOD_WLAN					=	MSECTOCOUNT(1000),			//!< LAN 接続状態タイマー起動周期
		PERIOD_VOLUME				=	MSECTOCOUNT(250),			//!< 音量タイマー起動周期
		PERIOD_PLAY					=	MSECTOCOUNT(500),			//!< 再生状態タイマー起動周期
		PERIOD_RECORD				=	MSECTOCOUNT(500),			//!< 録音状態タイマー起動周期
		PERIOD_CPUTEMP				=	MSECTOCOUNT(3000),			//!< CPU 温度起動周期
		WAITPLAY					=	MSECTOCOUNT(3000) / PERIOD_PLAY,
																	//!< 再生待ち
		ASCII_WIDTH					=	128,						//!< アスキー画像幅
		ASCII_HEIGHT				=	128,						//!< アスキー画像高さ
		DATE_LINE					=	0,							//!< 日付描画行
		DATE_XPOS					=	0,							//!< 日付描画ｘ座標
		TIME_LINE					=	0,							//!< 時間描画行
		TIME_XPOS					=	32,							//!< 時間描画ｘ座標
		BTCONNECT_LINE				=	0,							//!< Bluetooth アイコン描画行
		BTCONNECT_XPOS				=	64,							//!< Bluetooth アイコン描画ｘ座標
		BTCONNECT_ICON				=	0x90,						//!< Bluetooth アイコン
		WLANLQ_LINE					=	0,							//!< WLAN リンク品質アイコン描画行
		WLANLQ_XPOS					=	72,							//!< WLAN リンク品質アイコン描画ｘ座標
		WLANLQ_ICON					=	0x91,						//!< WLAN リンク品質アイコン
		VOLUME_LINE					=	0,							//!< 音量アイコン描画行
		VOLUME_XPOS					=	80,							//!< 音量アイコン描画ｘ座標
		VOLUME_ICON					=	0x97,						//!< 音量アイコン
		PLAY_LINE					=	0,							//!< 再生アイコン描画行
		PLAY_XPOS					=	88,							//!< 再生アイコン描画ｘ座標
		PLAY_ICON					=	0x9d,						//!< 再生アイコン
		RECORD_LINE					=	0,							//!< 録音アイコン描画行
		RECORD_XPOS					=	96,							//!< 録音アイコン描画ｘ座標
		RECORD_ICON					=	0x9f,						//!< 録音アイコン
		CPUTEMP_LINE				=	0,							//!< CPU 温度描画行
		CPUTEMP_XPOS				=	96,							//!< CPU 温度描画ｘ座標
		CPUTEMP_ICON				=	0x80,						//!< CPU 温度アイコン
		STATION_LINE				=	1,							//!< ラジオ局描画行
		STATION_XPOS				=	0,							//!< ラジオ局描画ｘ座標
		STATION_WIDTH				=	SSD1306_WIDTH,				//!< ラジオ局画像幅
		STATION_HEIGHT				=	40,							//!< ラジオ局画像高さ
		STATION_LINECOUNT			=	STATION_HEIGHT / SSD1306_LINEHEIGHT,
																	//!< ラジオ局描画行数
		STATION_ARROW				=	0x8e,						//!< ラジオ局矢印
		VOLUMEBAR_LINE				=	7,							//!< 音量バー描画行
		VOLUMEBAR_XPOS				=	2,							//!< 音量バー描画ｘ座標
		MENU_LINE					=	6,							//!< メニュー描画行
		MENU_XPOS					=	0,							//!< メニュー描画ｘ座標
		MENU_WIDTH					=	SSD1306_WIDTH,				//!< メニュー描画幅
		MENU_HEIGHT					=	16,							//!< メニュー描画高さ
		MENU_LINECOUNT				=	MENU_HEIGHT / SSD1306_LINEHEIGHT,
																	//!< メニュー描画行数
		MENU_LAYER_MAX				=	8,							//!< メニューレイヤー最大数
		MENUCURSOR_SPEED			=	2,							//!< メニューカーソル移動速度
	} ;

	//! 更新フラグ
	enum {
		UPDATE_DATE					=	1 <<  0,					//!< 日付
		UPDATE_TIME					=	1 <<  1,					//!< 日時
		UPDATE_BLUETOOTH			=	1 <<  2,					//!< Bluetooth 接続状態
		UPDATE_WLAN					=	1 <<  3,					//!< WLAN 接続状態
		UPDATE_VOLUME				=	1 <<  4,					//!< 音量
		UPDATE_PLAY					=	1 <<  5,					//!< 再生状態
		UPDATE_RECORD				=	1 <<  6,					//!< 再生状態
		UPDATE_CPUTEMP				=	1 <<  7,					//!< CPU 温度
		UPDATE_STATION				=	1 <<  8,					//!< ラジオ局
		UPDATE_VOLUMEBAR			=	1 <<  9,					//!< 音量バー
		UPDATE_MENU					=	1 << 10,					//!< メニュー
	} ;

	//! 状態フラグ
	enum {
		STATUS_DISPLAYOFF			=	1 << 0,						//!< 表示オフ
		STATUS_BLUETOOTH			=	1 << 1,						//!< Bluetooth 接続中
		STATUS_WLAN					=	1 << 2,						//!< WLAN 接続中
		STATUS_ENABLE				=	1 << 3,						//!< 再生有効
		STATUS_PLAY					=	1 << 4,						//!< 再生可能状態
		STATUS_RECORD				=	1 << 5,						//!< 録音状態
	} ;

	//! メイン状態
	enum {
		MAINSTATE_WAIT,												//!< 待ち
		MAINSTATE_SELECTMENU,										//!< メニュー選択
		MAINSTATE_SELECTSTATION,									//!< ラジオ局選択
		MAINSTATE_SELECTVOLUME,										//!< 音量選択
		MAINSTATE_SELECTRECORDING,									//!< 録音選択
		MAINSTATE_STOPRECORDING,									//!< 録音停止
		MAINSTATE_SELECTSHUTDOWN,									//!< シャットダウン選択
		MAINSTATE_SELECTREBOOT,										//!< 再起動選択
		MAINSTATE_MAX												//!< メイン処理最大数
	} ;

	//! タイマー ID
	enum {
		TIMERID_DATETIME,											//!< 日時
		TIMERID_BLUETOOTH,											//!< Bluetooth 接続状態
		TIMERID_WLAN,												//!< WLAN 接続状態
		TIMERID_VOLUME,												//!< 音量
		TIMERID_PLAY,												//!< 再生状態
		TIMERID_RECORD,												//!< 録音状態
		TIMERID_CPUTEMP,											//!< CPU 温度
		TIMERID_MAX													//!< 最大数
	} ;

	//! タイマー情報
	typedef struct _timerinfo {
		struct _timerinfo			*p_next ;						//!< 次のタイマー情報
		uint32_t					count ;							//!< 残りカウンタ
		TIMERPROC					proc ;							//!< タイマー処理
	} TIMERINFO ;

	//! ボタン状態
	enum {
		BTN_RED						=	1 << 0,						//!< 赤ボタン
		BTN_YELLOW					=	1 << 1,						//!< 黄ボタン
		BTN_BLUE					=	1 << 2,						//!< 青ボタン
		BTN_GREEN					=	1 << 3,						//!< 緑ボタン
	} ;

	//! ボタン情報
	typedef struct {
		uint8_t						norm ;							//!< 通常
		uint8_t						old ;							//!< 以前
		uint8_t						once ;							//!< １度のみ
		uint8_t						rept ;							//!< 繰り返し
		uint16_t					reptCount ;						//!< 繰り返しカウント
	} BUTTONINFO ;

	//! メニュー項目情報
	typedef struct {
		const char					*p_name ;						//!< 名前
		MENUITEMPROC				proc ;							//!< メニュー項目選択処理
		MENUINFO					*p_subMenuInfo ;				//!< サブメニュー情報
	} MENUITEMINFO ;

	//! メニュー情報
	struct _menuinfo {
		const MENUITEMINFO			*p_menuItemInfo ;				//!< メニュー項目情報
		int							menuItemCount ;					//!< メニュー項目数
		int							top ;							//!< メニュー上
		int							cursor ;						//!< カーソル位置
		int							add ;							//!< 加算値
	} ;


static void setMainProc( int state ) ;
static void menuSelectStation( void ) ;
static void menuVolume( void ) ;
static void menuRecording( void ) ;
static void menuBluetoothConnect( void ) ;
static void menuBluetoothDisconnect( void ) ;
static void menuShutdown( void ) ;
static void menuReboot( void ) ;

//! ラジオ局表示幅データ
static const uint8_t				m_stationWidth[]={
										128,
										124,
										115,
										100,
										 79,
										 55,
										 28,
//										  0,
									} ;

//! Bluetooth メニュー項目情報
static const MENUITEMINFO			m_bluetoothMenuItemInfo[]={
										{ "Connect",        menuBluetoothConnect,    0 },
										{ "Disconnect",     menuBluetoothDisconnect, 0 },
									} ;

//! Bluetooth メニュー情報
static MENUINFO						m_bluetoothMenuInfo={
										&m_bluetoothMenuItemInfo[0], countof( m_bluetoothMenuItemInfo ), 0, 0
									} ;

//! メインメニュー項目情報
static const MENUITEMINFO			m_mainMenuItemInfo[]={
										{ "Select Station", menuSelectStation, NULL                 },
										{ "Volume",         menuVolume,        NULL                 },
										{ "Recording",      menuRecording,     NULL                 },
										{ "Bluetooth",      NULL,              &m_bluetoothMenuInfo },
										{ "Shutdown",       menuShutdown,      NULL                 },
										{ "Reboot",         menuReboot,        NULL                 },
									} ;

//! メインメニュー情報
static MENUINFO						m_mainMenuInfo={
										&m_mainMenuItemInfo[0], countof( m_mainMenuItemInfo ), 0, 0
									} ;


static uint8_t						m_asciiData[256][8] ;			//!< アスキーデータ
static uint8_t						*m_p_stationData ;				//!< ラジオ局データ
static int							m_stationDataCount ;			//!< ラジオ局データ数
static SSD1306DATA					m_ssd1306data ;					//!< SSD1306 データ
static TIMERINFO					m_timerInfo[TIMERID_MAX] ;		//!< タイマー情報
static TIMERINFO					*m_p_timerInfoList ;			//!< タイマー情報リスト
static int							m_mainState ;					//!< メイン状態
static MAINPROC						m_mainProc ;					//!< メイン処理
static BUTTONINFO					m_buttonInfo ;					//!< ボタン情報
static uint8_t						m_keyButtonStateOld ;			//!< 以前のキーボタン状態
static bool							m_text ;						//!< テキストモードフラグ
static struct termios				m_termios ;						//!< 起動時の termios
static bool							m_exit ;						//!< メイン処理終了フラグ
static uint32_t						m_tickCount ;					//!< ティックカウント
static uint32_t						m_update ;						//!< 更新フラグ
static int							m_displayOffCount ;				//!< 表示オフダウンカウンタ
static int							m_contrast ;					//!< コントラスト
static int							m_contrastOld ;					//!< 以前のコントラスト
static uint32_t						m_status ;						//!< 状態フラグ
static struct tm					m_tm ;							//!< 日時
static int							m_btConnect ;					//!< Bluetooth 接続
static int							m_wlanLQ ;						//!< WLAN リンク品質
static int							m_wlanLQCode ;					//!< WLAN リンク品質文字コード
static int							m_volume ;						//!< 音量
static int							m_volumeCode ;					//!< 音量文字コード
static int							m_waitPlayCount ;				//!< 再生待ちカウンタ
static bool							m_play ;						//!< 再生状態
static bool							m_record ;						//!< 録音状態
static int							m_cpuTemp ;						//!< CPU 温度
static int							m_selectStation ;				//!< 選択ラジオ局
static int							m_fracStation ;					//!< ラジオ局選択時の移動端数
static int							m_addStation ;					//!< ラジオ局選択時の移動加算値
static MENUINFO						*m_p_menuInfo[MENU_LAYER_MAX] ;	//!< メニュー情報
static MENUINFO						*m_p_selectMenuInfo ;			//!< 選択メニュー情報
static int							m_menuInfoDepth ;				//!< メニュー情報の深さ
static int							m_accelMenuItem ;				//!< メニュー選択時の移動加速
static int							m_recordTime ;					//!< 録音時間
static int							m_longPressCount ;				//!< 長押しカウンタ


// -------------------------------------------------------------------------------- 
//!  シグナルハンドラ                                                               
//!                                                                                 
//!  @param [in] sig  シグナル番号                                                  
//!  @return          なし                                                          
// -------------------------------------------------------------------------------- 
static void signalHandler( int sig )
{
	m_exit = true ;
}


// -------------------------------------------------------------------------------- 
//!  256 色ビットマップから縦８ドットのビットデータへ変換                           
//!                                                                                 
//!  @param [in]  p_bi    ビットマップ情報                                          
//!  @param [out] p_data  変換データ格納アドレス                                    
//!  @return              なし                                                      
// -------------------------------------------------------------------------------- 
static void convImageData(BITMAPINFO *p_bi, uint8_t *p_data)
{
	uint8_t		*p_srce, *p_srceEnd, *p_dest ;
	uint32_t	*p_color ;
	int			y ;


	p_srce  = (uint8_t *)&p_bi->bmiColors[p_bi->bmiHeader.biClrUsed] ;
	p_dest  = &p_data[0] ;
	p_color = (uint32_t *)&p_bi->bmiColors[0] ;

	for(y=0; y<p_bi->bmiHeader.biHeight; y+=SSD1306_LINEHEIGHT) {
		p_srceEnd = &p_srce[p_bi->bmiHeader.biWidth] ;

		do {
			*p_dest++ = ((p_color[p_srce[p_bi->bmiHeader.biWidth*0]]!=0) << 0) |
			            ((p_color[p_srce[p_bi->bmiHeader.biWidth*1]]!=0) << 1) |
			            ((p_color[p_srce[p_bi->bmiHeader.biWidth*2]]!=0) << 2) |
			            ((p_color[p_srce[p_bi->bmiHeader.biWidth*3]]!=0) << 3) |
			            ((p_color[p_srce[p_bi->bmiHeader.biWidth*4]]!=0) << 4) |
			            ((p_color[p_srce[p_bi->bmiHeader.biWidth*5]]!=0) << 5) |
			            ((p_color[p_srce[p_bi->bmiHeader.biWidth*6]]!=0) << 6) |
			            ((p_color[p_srce[p_bi->bmiHeader.biWidth*7]]!=0) << 7) ;
		} while( ++p_srce != p_srceEnd ) ;

		p_srce += (p_bi->bmiHeader.biWidth * (SSD1306_LINEHEIGHT - 1)) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  アスキー画像データをロード                                                     
//!                                                                                 
//!  @param [in]  p_fileName  ファイル名                                            
//!  @param [out] p_data      ビットデータデータ格納アドレス                        
//!  @retval       0          正常終了                                              
//!  @retval      -1          異常終了                                              
// -------------------------------------------------------------------------------- 
static int loadAsciiImage(char const *p_fileName, uint8_t *p_data)
{
	BITMAPINFO	*p_bi ;
	int			result ;


	p_bi = bitmap_load( &p_fileName[0] ) ;
	if( p_bi == NULL ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return  -1 ;
	}

	result = -1 ;

	do {
		// 画像サイズ確認
		if( (p_bi->bmiHeader.biWidth!=ASCII_WIDTH) || (p_bi->bmiHeader.biHeight!=ASCII_HEIGHT) ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		convImageData(p_bi, &p_data[0]) ;

		// 正常終了
		result = 0 ;
	} while( 0 ) ;

	bitmap_free( p_bi ) ;

	return result ;
}


// -------------------------------------------------------------------------------- 
//!  ラジオ局画像データをロード                                                     
//!                                                                                 
//!  @param [in]  p_fileName  ファイル名                                            
//!  @param [out] pp_data     画像データアドレスの格納アドレス（要free）            
//!  @retval      -1          異常終了                                              
//!  @retval      上記以外    画像データ数                                          
// -------------------------------------------------------------------------------- 
static int loadStationImage(char const *p_fileName, uint8_t **pp_data)
{
	BITMAPINFO	*p_bi ;
	uint8_t		*p_data ;
	int			result ;


	p_bi = bitmap_load( &p_fileName[0] ) ;
	if( p_bi == NULL ) {
		return -1 ;
	}

	result = -1 ;
	p_data = NULL ;

	do {
		// 画像サイズ確認
		if( (p_bi->bmiHeader.biWidth!=STATION_WIDTH) || (p_bi->bmiHeader.biHeight%STATION_HEIGHT) ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		// 画像データ変換
		p_data = malloc( p_bi->bmiHeader.biWidth * (p_bi->bmiHeader.biHeight/SSD1306_LINEHEIGHT) ) ;
		if( p_data == NULL ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		convImageData(p_bi, &p_data[0]) ;

		// 正常終了
		result = p_bi->bmiHeader.biHeight / STATION_HEIGHT ;
	} while( 0 ) ;

	bitmap_free( p_bi ) ;

	*pp_data = &p_data[0] ;

	return result ;
}


// -------------------------------------------------------------------------------- 
//!  初期化前の状態に戻す                                                           
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void uninit( void )
{
	if( m_text != false ) {
		printf("\x1b[?25h") ;	// カーソル表示
		fflush( stdout ) ;
	}

	sound_close() ;
	ssd1306_uninit( &m_ssd1306data ) ;
	gpio_close() ;

	if( g_setupInfo.p_stationInfo != NULL ) {
		free( g_setupInfo.p_stationInfo ) ;
		g_setupInfo.p_stationInfo    = NULL ;
		g_setupInfo.stationInfoCount = 0 ;
	}

	if( m_p_stationData != NULL ) {
		free( m_p_stationData ) ;
		m_p_stationData    = NULL ;
		m_stationDataCount = 0 ;
	}
}


// -------------------------------------------------------------------------------- 
//!  初期化                                                                         
//!                                                                                 
//!  @retval  0  正常終了                                                           
//!  @retval -1  異常終了                                                           
// -------------------------------------------------------------------------------- 
static int init( void )
{
	struct termios	termios ;
	int				result ;


	result = -1 ;

	do {
		// アスキー画像ロード
		result = loadAsciiImage("data/ascii.bmp", &m_asciiData[0][0]) ;
		if( result < 0 ) {
			break ;
		}

		// ラジオ局画像ロード
		m_stationDataCount = loadStationImage("data/station.bmp", &m_p_stationData) ;
		if( m_stationDataCount < 0 ) {
			break ;
		}

		// 設定ファイルロード
		result = loadSetupFile( "setup.ini" ) ;
		if( result < 0 ) {
			return 1 ;
		}

		m_displayOffCount = MSECTOCOUNT( g_setupInfo.displayOffCount * 1000 ) ;
		m_contrast        = g_setupInfo.contrast ;
		m_contrastOld     = g_setupInfo.contrast ;

		// GPIO 開く
		result = gpio_open() ;
		if( result < 0 ) {
			break ;
		}

		// サウンド開く
		result = sound_open() ;
		if( result < 0 ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		if( m_text == false ) {		// SSD1306

			result = ssd1306_init(&m_ssd1306data, &g_setupInfo.i2cDev[0]) ;
			if( result < 0 ) {
				fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
				break ;
			}

			ssd1306_contrast(&m_ssd1306data, g_setupInfo.contrast) ;

		} else {					// テキスト

			memcpy(&termios, &m_termios, sizeof( m_termios )) ;
			termios.c_lflag    &= ~(ICANON|ECHO) ;
			termios.c_cc[VMIN]  = 0 ;
			termios.c_cc[VTIME] = 0 ;
			tcsetattr(STDIN_FILENO, TCSANOW, &termios) ;

			printf("\x1b[?25l"		// カーソル非表示
			       "\x1b[64B"		// カーソルを下へ64行移動
			       "\x1b[64A") ;	// カーソルを上へ64行移動
			fflush( stdout ) ;

		}

		// 正常終了
		return 0 ;
	} while( 0 ) ;

	uninit() ;

	return -1 ;
}


// -------------------------------------------------------------------------------- 
//!  タイマー情報リストへタイマー情報を取り付ける                                   
//!                                                                                 
//!  @param [in,out] p_timerInfo  タイマー情報                                      
//!  @param [in]     count        タイマー発生カウント                              
//!  @return                      なし                                              
// -------------------------------------------------------------------------------- 
static void attachTimerInfo(TIMERINFO *p_timerInfo, uint32_t count)
{
	TIMERINFO	**pp_timerInfoList, *p_nextTimerInfo ;


	pp_timerInfoList = &m_p_timerInfoList ;

	while( *pp_timerInfoList != NULL ) {
		p_nextTimerInfo = *pp_timerInfoList ;
		if( count < p_nextTimerInfo->count ) {
			p_nextTimerInfo->count -= count ;
			break ;
		}

		count           -= p_nextTimerInfo->count ;
		pp_timerInfoList = &p_nextTimerInfo->p_next ;
	}

	p_timerInfo->p_next = *pp_timerInfoList ;
	p_timerInfo->count  = count ;
	*pp_timerInfoList   = p_timerInfo ;
}


// -------------------------------------------------------------------------------- 
//!  タイマー処理を破棄                                                             
//!                                                                                 
//!  @param [in] timerId  タイマー識別子                                            
//!  @return              残りタイマーカウント                                      
// -------------------------------------------------------------------------------- 
static uint32_t killTimer( uint32_t timerId )
{
	TIMERINFO	*p_timerInfo, *p_nextTimerInfo, **pp_timerInfoList ;
	uint32_t	count ;


	p_timerInfo = &m_timerInfo[timerId] ;
	if( p_timerInfo->proc == NULL ) {
		return 0 ;
	}

	count = 0 ;

	pp_timerInfoList = &m_p_timerInfoList ;
	while( 1 ) {
		p_nextTimerInfo = *pp_timerInfoList ;
		if( p_nextTimerInfo == p_timerInfo ) {
			break ;
		}

		count += p_nextTimerInfo->count ;

		pp_timerInfoList = &p_nextTimerInfo->p_next ;
	}

	count += p_timerInfo->count ;

	*pp_timerInfoList = p_timerInfo->p_next ;

	if( p_timerInfo->p_next != NULL ) {
		p_timerInfo->p_next->count += p_timerInfo->count ;
	}

	p_timerInfo->proc = NULL ;

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  タイマー処理を追加                                                             
//!                                                                                 
//!  @param [in] timerId  タイマー識別子                                            
//!  @param [in] count    タイマー発生カウント                                      
//!  @param [in] proc     タイマー処理                                              
//!  @return              なし                                                      
// -------------------------------------------------------------------------------- 
static void addTimer(uint32_t timerId, uint32_t count, TIMERPROC proc)
{
	TIMERINFO	*p_timerInfo ;


	p_timerInfo = &m_timerInfo[timerId] ;
	if( p_timerInfo->proc != NULL ) {
		killTimer( timerId ) ;
	}

	p_timerInfo->proc = proc ;

	attachTimerInfo(p_timerInfo, count) ;
}


// -------------------------------------------------------------------------------- 
//!  タイマー処理                                                                   
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void timerProc( void )
{
	TIMERINFO	*p_timerInfo ;
	uint32_t	result ;


	p_timerInfo = m_p_timerInfoList ;
	if( p_timerInfo == NULL ) {
		return ;
	}

	if( --p_timerInfo->count ) {
		return ;
	}

	do {
		result = p_timerInfo->proc() ;

		m_p_timerInfoList = p_timerInfo->p_next ;

		if( result ) {
			attachTimerInfo(p_timerInfo, result) ;
		} else {
			p_timerInfo->proc = NULL ;
		}

		p_timerInfo = m_p_timerInfoList ;
		if( p_timerInfo == NULL ) {
			break ;
		}
	} while( !p_timerInfo->count ) ;
}


// -------------------------------------------------------------------------------- 
//!  キーのボタン状態を取得                                                         
//!                                                                                 
//!  @retval 0         ボタン押下なし                                               
//!  @retval 上記以外  ボタン状態                                                   
// -------------------------------------------------------------------------------- 
static uint8_t getKeyButton( void )
{
	char		code ;
	uint8_t		state ;
	int			result ;


	if( m_text == false ) {
		return 0 ;
	}

	result = read(STDIN_FILENO, &code, 1) ;
	if( result <= 0 ) {
		state               = m_keyButtonStateOld ;
		m_keyButtonStateOld = 0 ;
		return state ;
	}

	state = 0 ;

	switch( code ) {
		case 'z':
		{
			state = BTN_GREEN ;
			break ;
		}

		case 'x':
		{
			state = BTN_BLUE ;
			break ;
		}

		case 'c':
		{
			state = BTN_YELLOW ;
			break ;
		}

		case 'v':
		{
			state = BTN_RED ;
			break ;
		}

		default:
			break ;
	}

	m_keyButtonStateOld = state ;

	return state ;
}


// -------------------------------------------------------------------------------- 
//!  ボタン情報を更新                                                               
//!                                                                                 
//!  @param [in,out] p_buttonInfo  ボタン情報                                       
//!  @return                       なし                                             
// -------------------------------------------------------------------------------- 
static void updateButtonInfo( BUTTONINFO *p_buttonInfo )
{
	p_buttonInfo->old  = p_buttonInfo->norm ;
	p_buttonInfo->norm = gpio_button() | getKeyButton() ;
	p_buttonInfo->once = (p_buttonInfo->norm ^ p_buttonInfo->old) & p_buttonInfo->norm ;
	p_buttonInfo->rept = 0 ;

	if( p_buttonInfo->norm != p_buttonInfo->old ) {
		p_buttonInfo->rept      = p_buttonInfo->norm ;
		p_buttonInfo->reptCount = BUTTON_REPEAT ;
	} else {
		if( p_buttonInfo->reptCount ) {
			p_buttonInfo->reptCount-- ;
		} else {
			p_buttonInfo->rept = p_buttonInfo->norm ;
		}
	}
}


// -------------------------------------------------------------------------------- 
//!  コントラストをリセット                                                         
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void resetContrast( void )
{
	m_displayOffCount = MSECTOCOUNT( g_setupInfo.displayOffCount * 1000 ) ;
	m_contrast        = g_setupInfo.contrast ;
}


// -------------------------------------------------------------------------------- 
//!  コントラストを更新                                                             
//!                                                                                 
//!  @retval true   更新あり                                                        
//!  @retval false  更新なし                                                        
// -------------------------------------------------------------------------------- 
static bool updateContrast( void )
{
	int		diff ;


	if( m_text != false ) {
		return false ;
	}

	if( m_buttonInfo.norm ) {
		resetContrast() ;
	} else
	if( m_displayOffCount ) {
		m_displayOffCount-- ;
		if( !m_displayOffCount ) {
			m_status |= STATUS_DISPLAYOFF ;
			ssd1306_display(&m_ssd1306data, false) ;
		} else
		if( m_displayOffCount == CONTRAST_COUNT ) {
			m_contrast = 1 ;
		}
	}

	diff = m_contrast - m_contrastOld ;
	if( !diff ) {
		return false ;
	}

	if( diff >  10 ) {
		diff =  10 ;
	} else
	if( diff < -10 ) {
		diff = -10 ;
	}
	m_contrastOld += diff ;

	ssd1306_contrast(&m_ssd1306data, m_contrastOld) ;

	return true ;
}


// -------------------------------------------------------------------------------- 
//!  6x8 アスキー文字を描画                                                         
//!                                                                                 
//!  @param [in] line  行                                                           
//!  @param [in] x     ｘ座標                                                       
//!  @param [in] code  アスキーコード                                               
//!  @return           なし                                                         
// -------------------------------------------------------------------------------- 
static inline void drawAscii6x8(int line, int x, char code)
{
	ssd1306_drawImageLine(&m_ssd1306data, line, x, 6, 1, &m_asciiData[(uint8_t )code][0]) ;
}


// -------------------------------------------------------------------------------- 
//!  8x8 アスキー文字を描画                                                         
//!                                                                                 
//!  @param [in] line  行                                                           
//!  @param [in] x     ｘ座標                                                       
//!  @param [in] code  アスキーコード                                               
//!  @return           なし                                                         
// -------------------------------------------------------------------------------- 
static inline void drawAscii8x8(int line, int x, char code)
{
	ssd1306_drawImageLine(&m_ssd1306data, line, x, 8, 1, &m_asciiData[(uint8_t )code][0]) ;
}


// -------------------------------------------------------------------------------- 
//!  文字列を 6x8 アスキー文字で描画                                                
//!                                                                                 
//!  @param [in] line      行                                                       
//!  @param [in] x         ｘ座標                                                   
//!  @param [in] p_string  文字列                                                   
//!  @return               文字数                                                   
// -------------------------------------------------------------------------------- 
static inline int drawString6x8(int line, int x, const char *p_string)
{
	int		count ;


	count = 0 ;

	while( *p_string != '\0' ) {
		drawAscii6x8(line, x, *p_string++) ;
		x += 6 ;
		count++ ;
	}

	return count ;
}


// -------------------------------------------------------------------------------- 
//!  文字列を 6x8 アスキー文字で描画                                                
//!                                                                                 
//!  @param [in] line      行                                                       
//!  @param [in] x         ｘ座標                                                   
//!  @param [in] p_string  文字列                                                   
//!  @return               文字数                                                   
// -------------------------------------------------------------------------------- 
static inline int drawString8x8(int line, int x, const char *p_string)
{
	int		count ;


	count = 0 ;

	while( *p_string != '\0' ) {
		drawAscii8x8(line, x, *p_string++) ;
		x += 8 ;
		count++ ;
	}

	return count ;
}


// -------------------------------------------------------------------------------- 
//!  日時を更新                                                                     
//!                                                                                 
//!  @return 次タイマー発生カウント                                                 
// -------------------------------------------------------------------------------- 
static uint32_t updateDateTime( void )
{
	time_t		t ;
	struct tm	tm ;


	time( &t ) ;
	localtime_r(&t, &tm) ;

	// 月描画
	tm.tm_mon++ ;
	if( m_tm.tm_mon != tm.tm_mon ) {
		m_tm.tm_mon  = tm.tm_mon ;
		drawAscii6x8(DATE_LINE, DATE_XPOS+(0*6), '0'+(tm.tm_mon/10)) ;
		drawAscii6x8(DATE_LINE, DATE_XPOS+(1*6), '0'+(tm.tm_mon%10)) ;
		m_update |= UPDATE_DATE ;
	}

	// 日描画
	if( m_tm.tm_mday != tm.tm_mday ) {
		m_tm.tm_mday  = tm.tm_mday ;
		drawAscii6x8(DATE_LINE, DATE_XPOS+(3*6), '0'+(tm.tm_mday/10)) ;
		drawAscii6x8(DATE_LINE, DATE_XPOS+(4*6), '0'+(tm.tm_mday%10)) ;
		m_update |= UPDATE_DATE ;
	}

	// 時描画
	if( m_tm.tm_hour != tm.tm_hour ) {
		m_tm.tm_hour  = tm.tm_hour ;
		drawAscii6x8(TIME_LINE, TIME_XPOS+(0*6), '0'+(tm.tm_hour/10)) ;
		drawAscii6x8(TIME_LINE, TIME_XPOS+(1*6), '0'+(tm.tm_hour%10)) ;
		m_update |= UPDATE_TIME ;
	}

	// 分描画
	if( m_tm.tm_min != tm.tm_min ) {
		m_tm.tm_min  = tm.tm_min ;
		drawAscii6x8(TIME_LINE, TIME_XPOS+(3*6), '0'+(tm.tm_min/10)) ;
		drawAscii6x8(TIME_LINE, TIME_XPOS+(4*6), '0'+(tm.tm_min%10)) ;
		m_update |= UPDATE_TIME ;
	}

	return PERIOD_DATETIME ;
}


// -------------------------------------------------------------------------------- 
//!  Bluetooth を状態を更新                                                         
//!                                                                                 
//!  @return 次タイマー発生カウント                                                 
// -------------------------------------------------------------------------------- 
static uint32_t updateBluetooth( void )
{
	int		connect ;


	connect = (bluetooth_isConnect( &g_setupInfo.btAddr[0] ) == 1) ;
	if( m_btConnect == connect ) {
		return PERIOD_BLUETOOTH ;
	}

	m_btConnect = connect ;

	if( connect ) {		// 接続
		resetContrast() ;
		m_status       |=  STATUS_BLUETOOTH ;
		m_waitPlayCount = WAITPLAY ;
		drawAscii8x8(BTCONNECT_LINE, BTCONNECT_XPOS, BTCONNECT_ICON) ;
	} else {			// 切断
		m_status       &= ~(STATUS_BLUETOOTH|STATUS_ENABLE|STATUS_PLAY) ;
		m_waitPlayCount = 0 ;
		sound_stopPlay() ;
		drawAscii8x8(BTCONNECT_LINE, BTCONNECT_XPOS, ' ') ;
	}

	m_update |= UPDATE_BLUETOOTH ;

	return PERIOD_BLUETOOTH ;
}


// -------------------------------------------------------------------------------- 
//!  WLAN 状態を更新                                                                
//!                                                                                 
//!  @return 次タイマー発生カウント                                                 
// -------------------------------------------------------------------------------- 
static uint32_t updateWLan( void )
{
	uint8_t		code ;
	int			result ;


	result = wlan_getLinkQuality( &g_setupInfo.wlanIf[0] ) ;
	if( m_wlanLQ == result ) {
		return PERIOD_WLAN ;
	}

	m_wlanLQ = result ;

	if( result >= 0 ) {		// 接続
		if( !(m_status&STATUS_WLAN) ) {
			resetContrast() ;
			m_status       |=  STATUS_WLAN ;
			m_waitPlayCount = WAITPLAY ;
		}
	} else {				// 切断
		m_status       &= ~(STATUS_WLAN|STATUS_ENABLE|STATUS_PLAY) ;
		m_waitPlayCount = 0 ;
		sound_stopPlay() ;
	}

	if( result < 50 ) {
		code = WLANLQ_ICON +     (result >=  0) + (result >= 25) ;
	} else {
		code = WLANLQ_ICON + 3 + (result >= 75) + (result == 100) ;
	}

	if( m_wlanLQCode == code ) {
		return PERIOD_WLAN ;
	}

	m_wlanLQCode = code ;

	drawAscii8x8(WLANLQ_LINE, WLANLQ_XPOS, code) ;
	m_update |= UPDATE_WLAN ;

	return PERIOD_WLAN ;
}


// -------------------------------------------------------------------------------- 
//!  音量を更新                                                                     
//!                                                                                 
//!  @return 次タイマー発生カウント                                                 
// -------------------------------------------------------------------------------- 
static uint32_t updateVolume( void )
{
	static const uint8_t	threshold[]={
								  0,
								 24,
								 49,
								 74,
								 99,
//								100,
							} ;
	uint8_t					code ;
	int						index, volume ;


	volume = sound_getVolume() ;
	if( volume == -1 ) {
		return PERIOD_VOLUME ;
	}

	m_volume = volume ;

	for(index=0; index<countof( threshold ); index++) {
		if( volume <= threshold[index] ) {
			break ;
		}
	}
	code = VOLUME_ICON + index ;

	if( m_volumeCode != code ) {
		drawAscii8x8(VOLUME_LINE, VOLUME_XPOS, code) ;
		m_update |= UPDATE_VOLUME ;
	}

	return PERIOD_VOLUME ;
}


// -------------------------------------------------------------------------------- 
//!  再生状態を更新                                                                 
//!                                                                                 
//!  @return 次タイマー発生カウント                                                 
// -------------------------------------------------------------------------------- 
static uint32_t updatePlay( void )
{
	int		play ;


	play = sound_isPlay() ;
	if( play != false ) {
		m_status       |= STATUS_ENABLE ;
		m_waitPlayCount = 0 ;
	} else
	if( (m_status&(STATUS_BLUETOOTH|STATUS_WLAN)) == (STATUS_BLUETOOTH|STATUS_WLAN) ) {
		if( m_waitPlayCount ) {
			if( !--m_waitPlayCount ) {
				m_status |= STATUS_ENABLE ;
			}
		}
	} else {
		m_status &= ~STATUS_ENABLE ;
	}

	if( !(m_status&STATUS_ENABLE) ) {
		play = -1 ;
	}

	if( m_play == play ) {
		return PERIOD_PLAY ;
	}

	m_play = play ;

	if( play >= 0 ) {
		drawAscii8x8(PLAY_LINE, PLAY_XPOS, PLAY_ICON+play) ;
	} else {
		drawAscii8x8(PLAY_LINE, PLAY_XPOS, ' ') ;
	}

	m_update |= UPDATE_PLAY ;

	return PERIOD_PLAY ;
}


// -------------------------------------------------------------------------------- 
//!  録音状態を更新                                                                 
//!                                                                                 
//!  @return 次タイマー発生カウント                                                 
// -------------------------------------------------------------------------------- 
static uint32_t updateRecord( void )
{
	bool	record ;
	char	code ;


	record = sound_isRecord() ;
	if( m_record == record ) {
		return PERIOD_RECORD ;
	}

	m_record = record ;

	if( record != false ) {
		m_status |= STATUS_RECORD ;
		code      = RECORD_ICON ;
	} else {
		m_status &= ~STATUS_RECORD ;
		code      = ' ' ;
	}
	drawAscii8x8(RECORD_LINE, RECORD_XPOS, code) ;

	m_update |= UPDATE_RECORD ;

	return PERIOD_RECORD ;
}


// -------------------------------------------------------------------------------- 
//!  CPU 温度を取得                                                                 
//!                                                                                 
//!  @retval -128      異常終了                                                     
//!  @retval 上記以外  CPU 温度                                                     
// -------------------------------------------------------------------------------- 
static int getCPUTemp( void )
{
	FILE		*p_file ;
	size_t		size ;
	char		buff[1024] ;


	p_file = fopen(CPUTEMP_NAME, "rt") ;
	if( p_file == NULL ) {
		fprintf(stderr, "******** %s(%d) : can't open %s\n", __func__, __LINE__, CPUTEMP_NAME) ;
		return -128 ;
	}

	size = fread(&buff[0], 1, sizeof( buff ), p_file) ;

	fclose( p_file ) ;

	if( size < 2 ) {
		return -1 ;
	}

	return (int )strtol(&buff[0], NULL, 10) ;
}


// -------------------------------------------------------------------------------- 
//!  CPU 温度を更新                                                                 
//!                                                                                 
//!  @return 次タイマー発生カウント                                                 
// -------------------------------------------------------------------------------- 
static uint32_t updateCPUTemp( void )
{
	int		cpuTemp ;


	cpuTemp = getCPUTemp() ;
	if( m_cpuTemp == cpuTemp ) {
		return PERIOD_CPUTEMP ;
	}

	m_cpuTemp = cpuTemp ;

	if( cpuTemp >= 0 ) {	// 取得成功
		drawAscii6x8(CPUTEMP_LINE, CPUTEMP_XPOS+8+(0*6), '0'         +((cpuTemp/10000)%10)) ;
		drawAscii6x8(CPUTEMP_LINE, CPUTEMP_XPOS+8+(1*6), '0'         +((cpuTemp/1000 )%10)) ;
		drawAscii6x8(CPUTEMP_LINE, CPUTEMP_XPOS+8+(2*6), CPUTEMP_ICON+((cpuTemp/100  )%10)) ;
	} else {				// 取得失敗             
		drawAscii6x8(CPUTEMP_LINE, CPUTEMP_XPOS+8+(0*6), '-') ;
		drawAscii6x8(CPUTEMP_LINE, CPUTEMP_XPOS+8+(1*6), '-') ;
		drawAscii6x8(CPUTEMP_LINE, CPUTEMP_XPOS+8+(2*6), CPUTEMP_ICON+10) ;
	}

	drawAscii6x8(CPUTEMP_LINE, CPUTEMP_XPOS+8+(3*6), CPUTEMP_ICON+11) ;

	m_update |= UPDATE_CPUTEMP ;

	return PERIOD_CPUTEMP ;
}


// -------------------------------------------------------------------------------- 
//!  状態を描画                                                                     
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void drawStatus( void )
{
	memset(&m_tm, 0, sizeof( m_tm )) ;
	m_status        = 0 ;
	m_tm.tm_hour    = -1 ;
	m_tm.tm_min     = -1 ;
	m_btConnect     = 0 ;
	m_wlanLQ        = -2 ;
	m_wlanLQCode    = 0 ;
	m_volume        = -1 ;
	m_volumeCode    = 0 ;
	m_waitPlayCount = 0 ;
	m_play          = false ;
	m_record        = false ;
	m_cpuTemp       = -128 ;

	drawAscii6x8(DATE_LINE, DATE_XPOS+(2*6), '/') ;
	drawAscii6x8(TIME_LINE, TIME_XPOS+(2*6), ':') ;

	updateDateTime() ;
	updateBluetooth() ;
	updateWLan() ;
	updateVolume() ;
	updatePlay() ;
	updateRecord() ;
	updateCPUTemp() ;
}


// -------------------------------------------------------------------------------- 
//!  放送局を描画                                                                   
//!                                                                                 
//!  @param [in] focus   true:フォーカスあり                                        
//!                     false:フォーカスなし                                        
//!  @return            なし                                                        
// -------------------------------------------------------------------------------- 
static void drawStation( bool focus )
{
	uint8_t		*p_srce, *p_dest, *p_data[2] ;
	int			left, right, width, x, value, add ;


	left  = m_selectStation ;
	right = m_selectStation + 1 ;
	if( right >= g_setupInfo.stationInfoCount ) {
		right  = 0 ;
	}

	p_dest    = &m_ssd1306data.vram.b[0][STATION_LINE] ;
	p_data[0] = &m_p_stationData[left *(STATION_WIDTH*STATION_LINECOUNT)] ;
	p_data[1] = &m_p_stationData[right*(STATION_WIDTH*STATION_LINECOUNT)] ;

	x     = 0 ;
	width = m_stationWidth[m_fracStation] ;

	// 左
	if( width ) {
		value = 0 ;
		add   = (STATION_WIDTH * FIXP_ONE) / width ;

		for(; x<width; x++) {
			p_srce    = &p_data[0][value/FIXP_ONE] ;
			p_dest[0] = p_srce[STATION_WIDTH*0] ;
			p_dest[1] = p_srce[STATION_WIDTH*1] ;
			p_dest[2] = p_srce[STATION_WIDTH*2] ;
			p_dest[3] = p_srce[STATION_WIDTH*3] ;
			p_dest[4] = p_srce[STATION_WIDTH*4] ;
			p_dest   += 8 ;
			value    += add ;
		}
	}

	// 右
	if( width != STATION_WIDTH ) {
		value = 0 ;
		add   = (STATION_WIDTH * FIXP_ONE) / (STATION_WIDTH - width) ;

		for(; x<STATION_WIDTH; x++) {
			p_srce    = &p_data[1][value/FIXP_ONE] ;
			p_dest[0] = p_srce[STATION_WIDTH*0] ;
			p_dest[1] = p_srce[STATION_WIDTH*1] ;
			p_dest[2] = p_srce[STATION_WIDTH*2] ;
			p_dest[3] = p_srce[STATION_WIDTH*3] ;
			p_dest[4] = p_srce[STATION_WIDTH*4] ;
			p_dest   += 8 ;
			value    += add ;
		}
	}

	if( focus == false ) {
		ssd1306_grayLine(&m_ssd1306data, STATION_LINE, STATION_XPOS, STATION_WIDTH, STATION_LINECOUNT) ;
	}

	m_update |= UPDATE_STATION ;
}


// -------------------------------------------------------------------------------- 
//!  ラジオ局を更新                                                                 
//!                                                                                 
//!  @param [in] add    移動加算値                                                  
//!  @retval     true   更新あり                                                    
//!  @retval     false  更新なし                                                    
// -------------------------------------------------------------------------------- 
static bool updateStation( int add )
{
	if( !add ) {
		if( !m_addStation ) {
			return false ;
		}
		add = m_addStation ;
	} else {
		m_addStation = add ;
	}

	m_fracStation += add ;

	if( m_fracStation < 0 ) {
		m_fracStation = countof( m_stationWidth ) - 1 ;
		if( --m_selectStation < 0 ) {
			m_selectStation = g_setupInfo.stationInfoCount - 1 ;
		}
	} else
	if( m_fracStation >= countof( m_stationWidth ) ) {
		m_fracStation = 0 ;
		if( ++m_selectStation >= g_setupInfo.stationInfoCount ) {
			m_selectStation = 0 ;
		}
	}

	if( !m_fracStation ) {
		m_addStation = 0 ;
	}

	drawStation( true ) ;

	return true ;
}


// -------------------------------------------------------------------------------- 
//!  フォーカスを描画                                                               
//!                                                                                 
//!  @param [in] x      ｘ座標                                                      
//!  @param [in] y      ｙ座標                                                      
//!  @param [in] width  幅                                                          
// -------------------------------------------------------------------------------- 
static void drawFocus(int x, int y, int width)
{
	uint8_t		*p_data, *p_dataEnd ;
	uint16_t	data ;


	p_data    = &m_ssd1306data.vram.b[x][y/SSD1306_LINEHEIGHT] ;
	p_dataEnd = &m_ssd1306data.vram.b[x+width-1][y/SSD1306_LINEHEIGHT] ;

	y &= (SSD1306_LINEHEIGHT-1) ;

	if( !y ) {

		*p_data ^= 0x7e ;
		p_data  += SSD1306_LINE ;

		do {
			*p_data ^= 0xff ;
			p_data  += SSD1306_LINE ;
		} while( p_data != p_dataEnd ) ;

		*p_data ^= 0x7e ;

	} else {

		data       = 0x7e << y ;
		p_data[0] ^= (uint8_t )(data >> 0) ;
		p_data[1] ^= (uint8_t )(data >> 8) ;
		p_data    += SSD1306_LINE ;

		data = 0xff << y ;
		do {
			p_data[0] ^= (uint8_t )(data >> 0) ;
			p_data[1] ^= (uint8_t )(data >> 8) ;
			p_data    += SSD1306_LINE ;
		} while( p_data != p_dataEnd ) ;

		data       = 0x7e << y ;
		p_data[0] ^= (uint8_t )(data >> 0) ;
		p_data[1] ^= (uint8_t )(data >> 8) ;

	}
}


// -------------------------------------------------------------------------------- 
//!  行指定でフォーカスを描画                                                       
//!                                                                                 
//!  @param [in] line   行                                                          
//!  @param [in] x      ｘ座標                                                      
//!  @param [in] width  幅                                                          
// -------------------------------------------------------------------------------- 
static void drawFocusLine(int line, int x, int width)
{
	drawFocus(x, line*SSD1306_LINEHEIGHT, width) ;
}


// -------------------------------------------------------------------------------- 
//!  メニュー項目を描画                                                             
//!                                                                                 
//!  @param [out] p_buff    描画バッファアドレス                                    
//!  @param [out] p_string  文字列アドレス                                          
//!  @return                なし                                                    
// -------------------------------------------------------------------------------- 
static void drawMenuItem(uint8_t *p_buff, const char *p_string)
{
	uint8_t		*p_data ;
	int			width ;


	width  = (int )strlen( &p_string[0] ) * 6 ;
	p_buff = &p_buff[(MENU_WIDTH-width)/2] ;

	while( *p_string ) {
		p_data    = &m_asciiData[(uint8_t )*p_string++][0] ;
		p_buff[0] = p_data[0] ;
		p_buff[1] = p_data[1] ;
		p_buff[2] = p_data[2] ;
		p_buff[3] = p_data[3] ;
		p_buff[4] = p_data[4] ;
		p_buff[5] = p_data[5] ;
		p_buff   += 6 ;
	}
}


// -------------------------------------------------------------------------------- 
//!  メニューを描画                                                                 
//!                                                                                 
//!  @param [in] p_menuInfo  メニュー情報                                           
//!  @param [in] focus       フォーカス（true:あり, false:なし）                    
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static void drawMenu(const MENUINFO *p_menuInfo, bool focus)
{
	uint8_t		buff[MENU_LINECOUNT+1][MENU_WIDTH] ;
	uint32_t	data, y ;
	int			index, shift, item ;


	memset(&buff[0][0], 0, sizeof( buff )) ;

	item = p_menuInfo->top / SSD1306_LINEHEIGHT ;

	drawMenuItem(&buff[0][0], &p_menuInfo->p_menuItemInfo[item+0].p_name[0]) ;
	drawMenuItem(&buff[1][0], &p_menuInfo->p_menuItemInfo[item+1].p_name[0]) ;

	shift = p_menuInfo->top & (SSD1306_LINEHEIGHT-1) ;
	if( shift ) {
		drawMenuItem(&buff[2][0], &p_menuInfo->p_menuItemInfo[item+2].p_name[0]) ;

		for(index=0; index<MENU_WIDTH; index++) {
			data           = (((uint32_t )buff[0][index] <<  0) |
			                  ((uint32_t )buff[1][index] <<  8) |
			                  ((uint32_t )buff[2][index] << 16)) >> shift ;
			buff[0][index] = (uint8_t )(data >> 0) ;
			buff[1][index] = (uint8_t )(data >> 8) ;
		}
	}

	ssd1306_drawImageLine(&m_ssd1306data, MENU_LINE, MENU_XPOS, MENU_WIDTH, MENU_LINECOUNT, &buff[0][0]) ;

	if( focus != false ) {
		y = (MENU_LINE * SSD1306_LINEHEIGHT) + ((p_menuInfo->cursor - p_menuInfo->top) & (MENU_WIDTH - 1)) ;
		drawFocus(MENU_XPOS, y, MENU_WIDTH) ;
	}

	m_update |= UPDATE_MENU ;
}


// -------------------------------------------------------------------------------- 
//!  メニューをクリア                                                               
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void clearMenu( void )
{
	ssd1306_clearLine(&m_ssd1306data, MENU_LINE, MENU_XPOS, MENU_WIDTH, MENU_LINECOUNT) ;

	m_update |= UPDATE_MENU ;

}


// -------------------------------------------------------------------------------- 
//!  メインメニューを描画                                                           
//!                                                                                 
//!  @param [in] focus   true:フォーカスあり                                        
//!                     false:フォーカスなし                                        
//!  @return            なし                                                        
// -------------------------------------------------------------------------------- 
static void drawMainMenu( bool focus )
{
	drawMenu(m_p_menuInfo[m_menuInfoDepth], focus) ;
}


// -------------------------------------------------------------------------------- 
//!  メニューを更新                                                                 
//!                                                                                 
//!  @param [in,out] p_menuInfo  メニュー情報                                       
//!  @param [in]     add         メニューカーソル移動加算値                         
//!  @return                     なし                                               
// -------------------------------------------------------------------------------- 
static void updateMenu(MENUINFO *p_menuInfo, int add)
{
	if( !add ) {
		add = p_menuInfo->add ;
	} else {
		p_menuInfo->add = add ;
	}

	p_menuInfo->cursor += p_menuInfo->add ;

	if( add < 0 ) {		// 上へ

		if( p_menuInfo->cursor < 0 ) {
			p_menuInfo->cursor = 0 ;
		}

		if( p_menuInfo->top > p_menuInfo->cursor ) {
			p_menuInfo->top = p_menuInfo->cursor ;
		}

	} else
	if( add > 0 ) {		// 下へ

		if( p_menuInfo->cursor > ((p_menuInfo->menuItemCount-1)*SSD1306_LINEHEIGHT) ) {
			p_menuInfo->cursor = ((p_menuInfo->menuItemCount-1)*SSD1306_LINEHEIGHT) ;
		}

		if( p_menuInfo->top < (p_menuInfo->cursor-(MENU_HEIGHT-SSD1306_LINEHEIGHT)) ) {
			p_menuInfo->top = (p_menuInfo->cursor-(MENU_HEIGHT-SSD1306_LINEHEIGHT)) ;
		}

	}

	if( !(p_menuInfo->cursor&(SSD1306_LINEHEIGHT-1)) ) {
		p_menuInfo->add = 0 ;
	}

	drawMenu(p_menuInfo, true) ;
}


// -------------------------------------------------------------------------------- 
//!  音量バーを描画                                                                 
//!                                                                                 
//!  @param [in] volume  音量                                                       
//!  @return             なし                                                       
// -------------------------------------------------------------------------------- 
static void drawVolumeBar( int volume )
{
	char		buff[8] ;
	uint8_t		*p_data ;
	int			index ;

	// 数値
	if( !volume ) {
		drawString6x8(VOLUMEBAR_LINE, VOLUMEBAR_XPOS, "MIN") ;
	} else
	if( volume == 100 ) {
		drawString6x8(VOLUMEBAR_LINE, VOLUMEBAR_XPOS, "MAX") ;
	} else {
		sprintf(&buff[0], "%3d", volume) ;
		drawString6x8(VOLUMEBAR_LINE, VOLUMEBAR_XPOS, &buff[0]) ;
	}

	// バー
	p_data = &m_ssd1306data.vram.b[VOLUMEBAR_XPOS+20][VOLUMEBAR_LINE] ;

	p_data[0] = 0x7e ;
	p_data   += SSD1306_LINE ;

	p_data[0] = 0x42 ;
	p_data   += SSD1306_LINE ;

	for(index=0; index<volume; index++) {
		p_data[0] = 0x5a ;
		p_data   += SSD1306_LINE ;
	}

	for(; index<100; index++) {
		p_data[0] = 0x42 ;
		p_data   += SSD1306_LINE ;
	}

	p_data[0] = 0x42 ;
	p_data   += SSD1306_LINE ;

	p_data[0] = 0x7e ;

	m_update |= UPDATE_VOLUMEBAR ;
}


// -------------------------------------------------------------------------------- 
//!  録音時間を描画                                                                 
//!                                                                                 
//!  @param [in] recordTime  録音時間                                               
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
static void drawRecordTime( int recordTime )
{
	char	buff[8] ;


	sprintf(&buff[0], " %4d ", recordTime) ;
	drawString6x8(MENU_LINE+1, MENU_XPOS+(MENU_WIDTH-(6*6))/2, &buff[0]) ;
	drawFocusLine(MENU_LINE+1, MENU_XPOS+(MENU_WIDTH-(6*6))/2, 6*6) ;

	m_update |= UPDATE_MENU ;
}


// -------------------------------------------------------------------------------- 
//!  メニュー選択 - ラジオ局選択                                                    
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void menuSelectStation( void )
{
	ssd1306_grayLine(&m_ssd1306data, MENU_LINE, MENU_XPOS, MENU_WIDTH, MENU_LINECOUNT) ;

	setMainProc( MAINSTATE_SELECTSTATION ) ;
}


// -------------------------------------------------------------------------------- 
//!  メニュー選択 - 音量                                                            
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void menuVolume( void )
{
	clearMenu() ;
	drawVolumeBar( m_volume ) ;

	setMainProc( MAINSTATE_SELECTVOLUME ) ;
}


// -------------------------------------------------------------------------------- 
//!  メニュー選択 - 録音                                                            
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void menuRecording( void )
{
	if( !(m_status&STATUS_WLAN) ) {
		return ;
	}

	if( !(m_status&STATUS_RECORD) ) {	// 録音
		clearMenu() ;
		drawString6x8(MENU_LINE+0, MENU_XPOS+(MENU_WIDTH-(17*6))/2, "Record minutes(s)") ;
		drawRecordTime( m_recordTime ) ;
		setMainProc( MAINSTATE_SELECTRECORDING ) ;
	} else {							// 停止
		setMainProc( MAINSTATE_STOPRECORDING ) ;
	}
}


// -------------------------------------------------------------------------------- 
//!  メニュー選択 - Bluetooth 接続                                                  
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void menuBluetoothConnect( void )
{
	pid_t	pid ;
	char	*p_param[4] ;
	int		result ;


	pid = fork() ;
	if( pid ) {
		return ;
	}

	p_param[0] = "bluetoothctl" ;
	p_param[1] = "connect" ;
	p_param[2] = &g_setupInfo.btAddr[0] ;
	p_param[3] = NULL ;

	result = execvp(p_param[0], &p_param[0]) ;

	fprintf(stderr, "******** %s(%d) : %d - %d\n", __func__, __LINE__, result, errno) ;

	exit( 1 ) ;
}


// -------------------------------------------------------------------------------- 
//!  メニュー選択 - Bluetooth 接続                                                  
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void menuBluetoothDisconnect( void )
{
	pid_t	pid ;
	char	*p_param[4] ;
	int		result ;


	pid = fork() ;
	if( pid ) {
		return ;
	}

	p_param[0] = "bluetoothctl" ;
	p_param[1] = "disconnect" ;
	p_param[2] = &g_setupInfo.btAddr[0] ;
	p_param[3] = NULL ;

	result = execvp(p_param[0], &p_param[0]) ;

	fprintf(stderr, "******** %s(%d) : %d - %d\n", __func__, __LINE__, result, errno) ;

	exit( 1 ) ;
}


// -------------------------------------------------------------------------------- 
//!  メニュー選択 - シャットダウン選択                                              
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void menuShutdown( void )
{
	setMainProc( MAINSTATE_SELECTSHUTDOWN ) ;
}


// -------------------------------------------------------------------------------- 
//!  メニュー選択 - 再起動選択                                                      
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void menuReboot( void )
{
	setMainProc( MAINSTATE_SELECTREBOOT ) ;
}


// -------------------------------------------------------------------------------- 
//!  メイン処理 - 待ち                                                              
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void mainProcWait( void )
{
	switch( m_buttonInfo.once ) {
		case BTN_GREEN:		// メニュー表示
		case BTN_BLUE:
		{
			drawStation( false ) ;
			drawMainMenu( true ) ;

			setMainProc( MAINSTATE_SELECTMENU ) ;
			break ;
		}

		case BTN_YELLOW:	// 決定
		{
			if( (m_status&(STATUS_BLUETOOTH|STATUS_WLAN|STATUS_ENABLE)) == (STATUS_BLUETOOTH|STATUS_WLAN|STATUS_ENABLE) ) {
				sound_play( &g_setupInfo.p_stationInfo[m_selectStation] ) ;
			}
			break ;
		}

		case BTN_RED:		// キャンセル
		{
			sound_stopPlay() ;
			break ;
		}

		default:
			break ;
	}
}


// -------------------------------------------------------------------------------- 
//!  メイン処理 - メニュー選択                                                      
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void mainProcSelectMenu( void )
{
	const MENUITEMINFO	*p_menuItemInfo ;


	switch( m_buttonInfo.rept ) {
		case BTN_GREEN:		// カーソル上
		{
			if( m_p_selectMenuInfo->cursor ) {
				updateMenu(m_p_selectMenuInfo, -MENUCURSOR_SPEED) ;
				return ;
			}

			if( m_buttonInfo.once != BTN_GREEN ) {
				return ;
			}

			m_p_selectMenuInfo->top = (m_p_selectMenuInfo->menuItemCount * SSD1306_LINEHEIGHT) - MENU_HEIGHT ;
			if( m_p_selectMenuInfo->top < 0 ) {
				m_p_selectMenuInfo->top = 0 ;
			}

			m_p_selectMenuInfo->cursor = (m_p_selectMenuInfo->menuItemCount - 1) * SSD1306_LINEHEIGHT ;

			updateMenu(m_p_selectMenuInfo, 0) ;
			return ;
		}

		case BTN_BLUE:		// カーソル下
		{
			if( m_p_selectMenuInfo->cursor < ((m_p_selectMenuInfo->menuItemCount-1)*SSD1306_LINEHEIGHT) ) {
				updateMenu(m_p_selectMenuInfo, MENUCURSOR_SPEED) ;
				return ;
			}

			if( m_buttonInfo.once != BTN_BLUE ) {
				return ;
			}

			m_p_selectMenuInfo->top    = 0 ;
			m_p_selectMenuInfo->cursor = 0 ;

			updateMenu(m_p_selectMenuInfo, 0) ;
			return ;
		}

		default:
		{
			if( m_p_selectMenuInfo->add ) {
				updateMenu(m_p_selectMenuInfo, 0) ;
				return ;
			}
			break ;
		}
	}

	// ボタン
	switch( m_buttonInfo.once ) {
		case BTN_YELLOW:	// 決定
		{
			p_menuItemInfo = &m_p_selectMenuInfo->p_menuItemInfo[m_p_selectMenuInfo->cursor/8] ;

			// サブメニュー
			if( p_menuItemInfo->p_subMenuInfo != NULL ) {
				m_p_menuInfo[++m_menuInfoDepth] = p_menuItemInfo->p_subMenuInfo ;
				m_p_selectMenuInfo              = p_menuItemInfo->p_subMenuInfo ;
				drawMainMenu( true ) ;
			} else {
				p_menuItemInfo->proc() ;
			}
			break ;
		}

		case BTN_RED:		// キャンセル
		{
			// サブメニュー
			if( m_menuInfoDepth ) {
				m_p_selectMenuInfo = m_p_menuInfo[--m_menuInfoDepth] ;
				drawMainMenu( true ) ;
				break ;
			}

			drawStation( true ) ;
			clearMenu() ;

			setMainProc( MAINSTATE_WAIT ) ;
			break ;
		}
	}
}


// -------------------------------------------------------------------------------- 
//!  メイン処理 - ラジオ局選択                                                      
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void mainProcSelectStation( void )
{
	switch( m_buttonInfo.rept ) {
		case BTN_GREEN:		// カーソル左
		{
			updateStation( -1 ) ;
			break ;
		}

		case BTN_BLUE:		// カーソル右
		{
			updateStation( 1 ) ;
			break ;
		}

		case BTN_YELLOW:	// 決定
		{
			if( m_addStation || m_fracStation ) {
				break ;
			}

			if( (m_status&(STATUS_BLUETOOTH|STATUS_WLAN|STATUS_ENABLE)) == (STATUS_BLUETOOTH|STATUS_WLAN|STATUS_ENABLE) ) {
				sound_play( &g_setupInfo.p_stationInfo[m_selectStation] ) ;
			}
			break ;
		}

		case BTN_RED:		// キャンセル
		{
			if( m_addStation || m_fracStation ) {
				break ;
			}

			drawStation( false ) ;
			drawMainMenu( true ) ;

			setMainProc( MAINSTATE_SELECTMENU ) ;
			break ;
		}

		default:
		{
			if( updateStation( 0 ) != false ) {
				break ;
			}

			if( (m_tickCount/4) & 1 ) {
				break ;
			}

			drawStation( true ) ;

			// カーソル
			if( !(m_tickCount&8) ) {
				drawAscii6x8(STATION_LINE+2, STATION_XPOS              +2, STATION_ARROW+0) ;
			} else {         
				drawAscii6x8(STATION_LINE+2, STATION_XPOS+STATION_WIDTH-9, STATION_ARROW+1) ;
			}
			break ;
		}
	}
}


// -------------------------------------------------------------------------------- 
//!  メイン処理 - 音量選択                                                          
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void mainProcSelectVolume( void )
{
	switch( m_buttonInfo.rept ) {
		case BTN_GREEN:		// カーソル左
		{
			if( !m_volume ) {
				break ;
			}

			m_volume-- ;

			sound_setVolume( m_volume ) ;
			drawVolumeBar( m_volume ) ;
			break ;
		}

		case BTN_BLUE:		// カーソル右
		{
			if( m_volume == 100 ) {
				break ;
			}

			m_volume++ ;

			sound_setVolume( m_volume ) ;
			drawVolumeBar( m_volume ) ;
			break ;
		}

		case BTN_YELLOW:	// 決定
		{
			break ;
		}

		case BTN_RED:		// キャンセル
		{
			drawMainMenu( true ) ;

			setMainProc( MAINSTATE_SELECTMENU ) ;
			break ;
		}

		default:
		{
			if( m_update & UPDATE_VOLUME ) {
				drawVolumeBar( m_volume ) ;
			}
			break ;
		}
	}
}


// -------------------------------------------------------------------------------- 
//!  メイン処理 - 録音選択                                                          
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void mainProcSelectRecording( void )
{
	if( !m_buttonInfo.norm ) {
			m_accelMenuItem = 16 ;
	} else
	if( m_buttonInfo.rept ) {
		if( m_accelMenuItem < (8*16) ) {
			m_accelMenuItem++ ;
		}
	}

	switch( m_buttonInfo.rept ) {
		case BTN_GREEN:		// カーソル左
		{
			m_recordTime -= (m_accelMenuItem / 16) ;
			if( m_recordTime < 1 ) {
				m_recordTime = 1 ;
			}

			drawRecordTime( m_recordTime ) ;
			break ;
		}

		case BTN_BLUE:		// カーソル右
		{
			m_recordTime += (m_accelMenuItem / 16) ;
			if( m_recordTime > (24*60) ) {
				m_recordTime = (24*60) ;
			}

			drawRecordTime( m_recordTime ) ;
			break ;
		}

		case BTN_YELLOW:	// 決定
		{
			if( !(m_status&STATUS_WLAN) ) {
				break ;
			}

			sound_record(&g_setupInfo.p_stationInfo[m_selectStation], m_recordTime) ;

			drawStation( false ) ;
			drawMainMenu( true ) ;

			setMainProc( MAINSTATE_SELECTMENU ) ;
			break ;
		}

		case BTN_RED:		// キャンセル
		{
			drawStation( false ) ;
			drawMainMenu( true ) ;

			setMainProc( MAINSTATE_SELECTMENU ) ;
			break ;
		}

		default:
		{
			break ;
		}
	}
}


// -------------------------------------------------------------------------------- 
//!  メイン処理 - 長押し                                                            
//!                                                                                 
//!  @return なし                                                                   
// -------------------------------------------------------------------------------- 
static void mainProcLongPress( void )
{
	int		index ;

	// キャンセル
	if( !(m_buttonInfo.norm&BTN_YELLOW) ) {
		m_longPressCount = 0 ;

		drawMainMenu( true ) ;
		setMainProc( MAINSTATE_SELECTMENU ) ;
		return ;
	}

	m_longPressCount += 2 ;
	if( m_longPressCount < (SSD1306_WIDTH/2) ) {
		drawMainMenu( false ) ;
		drawFocusLine(MENU_LINE+((m_p_selectMenuInfo->cursor-m_p_selectMenuInfo->top)/SSD1306_LINEHEIGHT), MENU_XPOS+m_longPressCount, MENU_XPOS+MENU_WIDTH-(m_longPressCount*2)) ;
		return ;
	}

	m_longPressCount = 0 ;

	switch( m_mainState ) {
		case MAINSTATE_STOPRECORDING:
		{
			drawMainMenu( true ) ;

			sound_stopRecord() ;

			setMainProc( MAINSTATE_SELECTMENU ) ;
			break ;
		}

		case MAINSTATE_SELECTSHUTDOWN:
		case MAINSTATE_SELECTREBOOT:
		{
			drawMainMenu( false ) ;

			// メッセージ表示
			for(index=0; index<3; index++) {
				ssd1306_clear( &m_ssd1306data ) ;
				ssd1306_update( &m_ssd1306data ) ;

				msleep( 500 ) ;

				if( m_mainState == MAINSTATE_SELECTSHUTDOWN ) {
					drawString6x8(7, 2, "Shutting down...") ;
				} else {
					drawString6x8(7, 2, "Rebooting...") ;
				}

				ssd1306_update( &m_ssd1306data ) ;

				msleep( 500 ) ;
			}

			uninit() ;

			sync() ;

			if( m_mainState == MAINSTATE_SELECTSHUTDOWN ) {
				system( "sudo shutdown -h now" ) ;
			} else {
				system( "sudo reboot" ) ;
			}

			exit( 0 ) ;
			break ;
		}

		default:
			break ;
	}
}


// -------------------------------------------------------------------------------- 
//!  メイン処理を設定                                                               
//!                                                                                 
//!  @param [in] state  メイン処理状態                                              
//!  @return            なし                                                        
// -------------------------------------------------------------------------------- 
static void setMainProc( int state )
{
	static const MAINPROC	mainProc[]={
								mainProcWait,				// MAINSTATE_WAIT
								mainProcSelectMenu,			// MAINSTATE_SELECTMENU
								mainProcSelectStation,		// MAINSTATE_SELECTSTATION
								mainProcSelectVolume,		// MAINSTATE_SELECTVOLUME
								mainProcSelectRecording,	// MAINSTATE_SELECTRECORDING
								mainProcLongPress,			// MAINSTATE_STOPRECORDING
								mainProcLongPress,			// MAINSTATE_SELECTSHUTDOWN
								mainProcLongPress,			// MAINSTATE_SELECTREBOOT
							} ;


	m_mainState = state ;
	m_mainProc  = mainProc[state] ;
}


// -------------------------------------------------------------------------------- 
//!  テキスト画面を更新                                                             
//!                                                                                 
//!  @param [in] p_data  SSD1306 データ                                             
//!  @return             なし                                                       
// -------------------------------------------------------------------------------- 
static void updateTextScreen( SSD1306DATA *p_data )
{
	char		buff[SSD1306_HEIGHT][SSD1306_WIDTH+1], *p_buff ;
	uint8_t		data ;
	int			x, y ;


	memset(&buff[0][0], ' ', sizeof( buff )) ;

	for(y=0; y<SSD1306_HEIGHT-1; y++) {
		buff[y][SSD1306_WIDTH] = '\n' ;
	}
	buff[y][SSD1306_WIDTH] = '\0' ;

	for(x=0; x<SSD1306_WIDTH; x++) {
		for(y=0; y<SSD1306_HEIGHT; y+=SSD1306_LINEHEIGHT) {
			p_buff = &buff[y][x] ;
			data   = p_data->vram.b[x][y/SSD1306_LINEHEIGHT] ;

			while( data ) {
				if( data & 1 ) {
					*p_buff = '#' ;
				}
				p_buff += countof( buff[0] ) ;

				data >>= 1 ;
			}
		}
	}

	printf("%s"
	       "\x1b[64A\r", &buff[0][0]) ;
	fflush( stdout ) ;
}


// -------------------------------------------------------------------------------- 
//!  表示オフ中処理                                                                 
//!                                                                                 
//!  @retval true   正常終了                                                        
//!  @retval false  処理終了                                                        
// -------------------------------------------------------------------------------- 
static bool displayOffProc( void )
{
	while( 1 ) {
		if( m_exit != false ) {
			return false ;
		}

		updateButtonInfo( &m_buttonInfo ) ;
		if( m_buttonInfo.norm ) {
			break ;
		}

		msleep( DISPLAYOFF_WAIT ) ;
	}

	setMainProc( MAINSTATE_WAIT ) ;

	ssd1306_clear( &m_ssd1306data ) ;

	drawStatus() ;
	drawStation( true ) ;

	ssd1306_display(&m_ssd1306data, true) ;

	m_status &= ~STATUS_DISPLAYOFF ;

	resetContrast() ;

	return true ;
}


// -------------------------------------------------------------------------------- 
//!  メイン                                                                         
//!                                                                                 
//!  @param [in] argc  引数の数                                                     
//!  @param [in] argv  引数                                                         
//!  @retval     0     正常終了                                                     
//!  @retval     1     異常終了                                                     
// -------------------------------------------------------------------------------- 
int main(int argc, char *argv[])
{
	uint32_t	base, current ;
	int32_t		diff ;
	int			result ;

	// 初期化
	memset(&g_setupInfo,  0, sizeof( g_setupInfo )) ;
	memset(&m_timerInfo,  0, sizeof( m_timerInfo )) ;
	memset(&m_buttonInfo, 0, sizeof( m_buttonInfo )) ;
	m_p_stationData     = NULL ;
	m_stationDataCount  = 0 ;
	m_ssd1306data.fd    = -1 ;
	m_p_timerInfoList   = NULL ;
	m_text              = false ;
	m_keyButtonStateOld = 0 ;
	m_exit              = false ;
	m_tickCount         = 0 ;
	m_update            = 0 ;
	m_selectStation     = 0 ;
	m_fracStation       = 0 ;
	m_addStation        = 0 ;
	m_p_menuInfo[0]     = &m_mainMenuInfo ;
	m_p_selectMenuInfo  = &m_mainMenuInfo ;
	m_menuInfoDepth     = 0 ;
	m_accelMenuItem     = 0 ;
	m_recordTime        = 60 ;
	m_longPressCount    = 0 ;

	setMainProc( MAINSTATE_WAIT ) ;

	if( argc >= 2 ) {
		if( !strcmp(argv[1], "-t") ) {
			m_text = true ;
		}
	}

	setlocale(LC_ALL, "") ;

	setCurrentDir( argv[0] ) ;

	tcgetattr(STDIN_FILENO, &m_termios) ;

	// 初期化
	result = init() ;
	if( result < 0 ) {
		return 1 ;
	}

	signal(SIGINT, signalHandler) ;

	addTimer(TIMERID_DATETIME,  PHASE_DATETIME,  updateDateTime) ;
	addTimer(TIMERID_BLUETOOTH, PHASE_BLUETOOTH, updateBluetooth) ;
	addTimer(TIMERID_WLAN,      PHASE_WLAN,      updateWLan) ;
	addTimer(TIMERID_VOLUME,    PHASE_VOLUME,    updateVolume) ;
	addTimer(TIMERID_PLAY,      PHASE_PLAY,      updatePlay) ;
	addTimer(TIMERID_RECORD,    PHASE_RECORD,    updateRecord) ;
	addTimer(TIMERID_CPUTEMP,   PHASE_CPUTEMP,   updateCPUTemp) ;

	drawStatus() ;
	drawStation( true ) ;

	ssd1306_update( &m_ssd1306data ) ;

	base = getCurrentTime() ;

	while( m_exit == false ) {

		// 表示オフ
		if( m_status & STATUS_DISPLAYOFF ) {
			if( displayOffProc() == false ) {
				break ;
			}
		}

		// タイミング調整
		base   += PERIOD_MS ;
		current = getCurrentTime() ;
		diff    = (int32_t )(base - current) ;
		if( diff > 0 ) {
			msleep( diff ) ;
		} else {
			base = current ;
		}

		m_tickCount++ ;

		// ボタン情報更新
		updateButtonInfo( &m_buttonInfo ) ;

		// タイマー処理
		timerProc() ;

		// メイン処理
		if( updateContrast() == false ) {
			m_mainProc() ;
		}

		// 画面更新
		if( m_update ) {
			m_update = 0 ;

			if( m_text == false ) {		// SSD1306
				ssd1306_update( &m_ssd1306data ) ;
			} else {					// テキスト
				updateTextScreen( &m_ssd1306data ) ;
			}
		}
	}

	uninit() ;

	tcsetattr(STDIN_FILENO, TCSANOW, &m_termios) ;

	return 0 ;
}
