// -------------------------------------------------------------------------------- 
//!  @file  sound.c                                                                 
//!  サウンド処理のソースファイル                                                   
//                                                                                  
//   Copyright (C) 2024 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <limits.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <alsa/asoundlib.h>
#include "common.h"
#include "sound.h"


static snd_mixer_t					*m_p_mixer=NULL ;				//!< ミキサーのハンドル
static snd_mixer_elem_t				*m_p_elem ;						//!< ミキサー要素のハンドル
static long							m_min ;							//!< 音量の最小値
static long							m_max ;							//!< 音量の最大値
static long							m_range ;						//!< 音量の範囲
static int							m_volume ;						//!< 音量
static pid_t						m_pid_play=-1 ;					//!< 再生プロセス ID
static pid_t						m_pid_record=-1 ;				//!< 録音プロセス ID


// -------------------------------------------------------------------------------- 
//!  開く                                                                           
//!                                                                                 
//!  @retval  0  正常終了                                                           
//!  @retval -1  異常終了                                                           
// -------------------------------------------------------------------------------- 
int sound_open( void )
{
	snd_mixer_t				*p_mixer ;
	snd_mixer_selem_id_t	*p_selemId ;
	snd_mixer_elem_t		*p_elem ;
	int						result ;
	long					min, max, volume ;


	if( m_p_mixer != NULL ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	result = snd_mixer_open(&p_mixer, 0) ;
	if( result < 0 ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	p_selemId = NULL ;

	do {
		result = snd_mixer_attach(p_mixer, &g_setupInfo.mixer[0]) ;
		if( result < 0 ) {
			fprintf(stderr, "******** %s(%d) : %s\n", __func__, __LINE__, snd_strerror( result )) ;
			break ;
		}

		result = snd_mixer_selem_register(p_mixer, NULL, NULL) ;
		if( result < 0 ) {
			fprintf(stderr, "******** %s(%d) : %s\n", __func__, __LINE__, snd_strerror( result )) ;
			break ;
		}

		result = snd_mixer_load( p_mixer ) ;
		if( result < 0 ) {
			fprintf(stderr, "******** %s(%d) : %s\n", __func__, __LINE__, snd_strerror( result )) ;
			break ;
		}

		snd_mixer_selem_id_alloca( &p_selemId ) ;
		if( p_selemId == NULL ) {
			result = -1 ;
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		snd_mixer_selem_id_set_index(p_selemId, 0) ;
		snd_mixer_selem_id_set_name(p_selemId, &g_setupInfo.selem[0]) ;

		p_elem = snd_mixer_find_selem(p_mixer, p_selemId) ;
		if( p_elem == NULL ) {
			result = -1 ;
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		result = snd_mixer_selem_get_playback_volume_range(p_elem, &min, &max) ;
		if( result < 0 ) {
			fprintf(stderr, "******** %s(%d) : %s\n", __func__, __LINE__, snd_strerror( result )) ;
			break ;
		}

		result = snd_mixer_selem_get_playback_volume(p_elem, SND_MIXER_SCHN_MONO, &volume) ;
		if( result < 0 ) {
			fprintf(stderr, "******** %s(%d) : %s\n", __func__, __LINE__, snd_strerror( result )) ;
			break ;
		}

		// 正常終了
		m_p_mixer    = p_mixer ;
		m_p_elem     = p_elem ;
		m_min        = min ;
		m_max        = max ;
		m_range      = (max - min) + 1 ;
		m_volume     = sound_getVolume() ;
		m_pid_play   = -1 ;

		return 0 ;
	} while( 0 ) ;

	snd_mixer_close( p_mixer ) ;

	return -1 ;
}


// -------------------------------------------------------------------------------- 
//!  閉じる                                                                         
//!                                                                                 
//!  @retval  0  正常終了                                                           
//!  @retval -1  異常終了                                                           
// -------------------------------------------------------------------------------- 
int sound_close( void )
{
	if( m_p_mixer == NULL ) {
		return -1 ;
	}

	sound_stopPlay() ;
	sound_stopRecord() ;

	snd_mixer_close( m_p_mixer ) ;

	m_p_mixer = NULL ;

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  音量を取得                                                                     
//!                                                                                 
//!  @retval 0-100  音量                                                            
//!  @retval -1     異常終了                                                        
// -------------------------------------------------------------------------------- 
int sound_getVolume( void )
{
	int		result ;
	long	volume ;


	if( m_p_mixer == NULL ) {
		return -1 ;
	}

	snd_mixer_handle_events( m_p_mixer ) ;

	result = snd_mixer_selem_get_playback_volume(m_p_elem, SND_MIXER_SCHN_MONO, &volume) ;
	if( result < 0 ) {
		fprintf(stderr, "******** %s(%d) : %s\n", __func__, __LINE__, snd_strerror( result )) ;
		return -1 ;
	}

	return (((volume-m_min)*100+(m_range/2-1))/m_range) ;
}


// -------------------------------------------------------------------------------- 
//!  音量を設定                                                                     
//!                                                                                 
//!  @retval 0-100  音量                                                            
//!  @retval -1     異常終了                                                        
// -------------------------------------------------------------------------------- 
int sound_setVolume( uint8_t volume )
{
	int		result ;
	long	value ;


	if( m_p_mixer == NULL ) {
		return -1 ;
	}

	if( !volume ) {
		value = m_min ;
	} else
	if( volume == 100 ) {
		value = m_max ;
	} else {
		value = ((long )volume * m_range + (m_range/2-1)) / 100 + m_min ;
	}

	result = snd_mixer_selem_set_playback_volume_all(m_p_elem, value) ;
	if( result < 0 ) {
		fprintf(stderr, "******** %s(%d) : %s\n", __func__, __LINE__, snd_strerror( result )) ;
		return -1 ;
	}

	m_volume = volume ;

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  再生                                                                           
//!                                                                                 
//!  @param [in] p_stationInfo  ラジオ局情報                                        
//!  @retval      0             正常終了                                            
//!  @retval     -1             異常終了                                            
// -------------------------------------------------------------------------------- 
int sound_play( STATIONINFO *p_stationInfo )
{
	char	*p_param[6] ;
	int		result, volume ;


	if( m_p_mixer == NULL ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	sound_stopPlay() ;

	// 再オープン
	snd_mixer_close( m_p_mixer ) ;
	m_p_mixer = NULL ;

	volume = m_volume ;

	sound_open() ;

	sound_setVolume( volume ) ;

	m_pid_play = fork() ;
	if( m_pid_play < 0 ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	if( m_pid_play ) {
		return 0 ;
	}

	setsid() ;

	p_param[0] = &g_setupInfo.script.play[0] ;
	p_param[1] = "-t" ;
	p_param[2] = &p_stationInfo->type[0] ;
	p_param[3] = "-s" ;
	p_param[4] = &p_stationInfo->id[0] ;
	p_param[5] = NULL ;
	result     = execvp(p_param[0], &p_param[0]) ;

	fprintf(stderr, "******** %s(%d) : %d(%d)\n", __func__, __LINE__, result, errno) ;

	exit( 0 ) ;
}


// -------------------------------------------------------------------------------- 
//!  再生を停止                                                                     
//!                                                                                 
//!  @retval  0  正常終了                                                           
//!  @retval -1  再生中でない                                                       
// -------------------------------------------------------------------------------- 
int sound_stopPlay( void )
{
	int		status ;


	if( m_pid_play == -1 ) {
		return -1 ;
	}

	if( !killpg(m_pid_play, SIGKILL) ) {
		if( waitpid(m_pid_play, &status, 0) < 0 ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		}
	}

	m_pid_play = -1 ;

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  再生中か調べる                                                                 
//!                                                                                 
//!  @retval true   実行中                                                          
//!  @retval false  停止中                                                          
// -------------------------------------------------------------------------------- 
bool sound_isPlay( void )
{
	int		status ;


	if( m_pid_play == -1 ) {
		return false ;
	}

	if( waitpid(m_pid_play, &status, WNOHANG) < 0 ) {
		m_pid_play = -1 ;
		return false ;
	}

	return true ;
}


// -------------------------------------------------------------------------------- 
//!  録音                                                                           
//!                                                                                 
//!  @param [in] p_stationInfo  ラジオ局情報                                        
//!  @param [in] minute         録音時間                                            
//!  @retval      0             正常終了                                            
//!  @retval     -1             異常終了                                            
// -------------------------------------------------------------------------------- 
int sound_record(STATIONINFO *p_stationInfo, int minute)
{
	char	path[PATH_MAX*2], dir[PATH_MAX], param[8], *p_param[8] ;
	int		result ;


	m_pid_record = fork() ;
	if( m_pid_record < 0 ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	if( m_pid_record ) {
		return 0 ;
	}

	getcwd(&dir[0], sizeof( dir )) ;
	snprintf(&path[0], sizeof( path ), "%s/%s", &dir[0], &g_setupInfo.script.record[0]) ;

	chdir( &g_setupInfo.outputRecord[0] ) ;

	sprintf(&param[0], "%d", minute) ;

	setsid() ;

	p_param[0] = &path[0] ;
	p_param[1] = "-t" ;
	p_param[2] = &p_stationInfo->type[0] ;
	p_param[3] = "-s" ;
	p_param[4] = &p_stationInfo->id[0] ;
	p_param[5] = "-d" ;
	p_param[6] = &param[0] ;
	p_param[7] = NULL ;
	result     = execvp(p_param[0], &p_param[0]) ;

	fprintf(stderr, "******** %s(%d) : %d(%d)\n", __func__, __LINE__, result, errno) ;

	exit( 0 ) ;
}


// -------------------------------------------------------------------------------- 
//!  録音を停止                                                                     
//!                                                                                 
//!  @retval  0  正常終了                                                           
//!  @retval -1  録音中でない                                                       
// -------------------------------------------------------------------------------- 
int sound_stopRecord( void )
{
	int		status ;


	if( m_pid_record == -1 ) {
		return -1 ;
	}

	if( !killpg(m_pid_record, SIGKILL) ) {
		if( waitpid(m_pid_record, &status, 0) < 0 ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		}
	}

	m_pid_record = -1 ;

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  録音中か調べる                                                                 
//!                                                                                 
//!  @retval true   実行中                                                          
//!  @retval false  停止中                                                          
// -------------------------------------------------------------------------------- 
bool sound_isRecord( void )
{
	int		status ;


	if( m_pid_record == -1 ) {
		return false ;
	}

	if( waitpid(m_pid_record, &status, WNOHANG) < 0 ) {
		m_pid_record = -1 ;
		return false ;
	}

	return true ;
}
