// -------------------------------------------------------------------------------- 
//!  @file  ssd1306.c                                                               
//!  SSD1306 制御のソースファイル                                                   
//                                                                                  
//   Copyright (C) 2024 lsluk, All rights reserved.                                 
//   http://vdmgr.g2.xrea.com/index.html                                            
// -------------------------------------------------------------------------------- 
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>
#include <fcntl.h>
#include "ssd1306.h"


// -------------------------------------------------------------------------------- 
//!  初期化                                                                         
//!                                                                                 
//!  @param [out] p_data  SSD1306 データ                                            
//!  @param [in]  p_name  デバイスファイル名                                        
//!  @retval       0      正常終了                                                  
//!  @retval      -1      異常終了                                                  
// -------------------------------------------------------------------------------- 
int ssd1306_init(SSD1306DATA *p_data, const char *p_name)
{
	static const uint8_t	command[]={
								3, 0x00,0xa8,0x3f,			// Set Multiplex Ratio
								3, 0x00,0xd3,0x00,			// Set Display Offset
								2, 0x00,0x40,				// Set Display Start Line
								2, 0x00,0xa1,				// Set Segment Re-map
								2, 0x00,0xc8,				// Set COM Output Scan Direction
								3, 0x00,0xda,0x12,			// Set COM Pins Hardware Configuration
								3, 0x00,0x81,0x7f,			// Set Contrast Control
								2, 0x00,0xa4,				// Entire Display ON
								2, 0x00,0xa6,				// Set Normal/Inverse Display
								3, 0x00,0xd5,0x80,			// Set Display Clock Divide Ratio/Oscillator Frequency
								3, 0x00,0x8d,0x14,			// Charge Pump Setting
								3, 0x00,0x20,0x01,			// Set Memory Addressing Mode
								4, 0x00,0x22,0x00,0x07,		// Set Page Address
								4, 0x00,0x21,0x00,0x7f,		// Set Column Address
								0,							// 画面クリア
								2, 0x00,0xaf				// Set Display ON/OFF
							} ;
	uint8_t					length ;
	const uint8_t			*p_command ;
	int						result, fd ;


	memset(p_data, 0, sizeof( SSD1306DATA )) ;
	p_data->fd      = -1 ;
	p_data->control = 0x40 ;

	fd = open(&p_name[0], O_RDWR) ;
	if( fd < 0 ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	do {
		result = ioctl(fd, I2C_SLAVE, SSD1306_SLAVE_ADDR) ;
		if( result < 0 ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		// 初期化
		p_command = &command[0] ;

		while( 1 ) {
			length = *p_command++ ;
			if( !length ) {
				break ;
			}

			result = write(fd, &p_command[0], length) ;
			if( result != length ) {
				fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
				break ;
			}

			p_command += length ;
		}
		if( length ) {
			break ;
		}

		// 画面クリア
		memset(&p_data->vram.b[0][0], 0, sizeof( p_data->vram.b )) ;
		result        = write(fd, &p_data->control, 1+sizeof( p_data->vram.b )) ;
		if( result != (1+sizeof( p_data->vram.b )) ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		length = *p_command++ ;

		// 表示
		result = write(fd, &p_command[0], length) ;
		if( result != length ) {
			fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
			break ;
		}

		p_data->fd = fd ;

		// 正常終了
		return 0 ;
	} while( 0 ) ;

	close( fd ) ;

	return -1 ;
}


// -------------------------------------------------------------------------------- 
//!  初期化前の状態に戻す                                                           
//!                                                                                 
//!  @param [in,out] p_data  SSD1306 データ                                         
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
void ssd1306_uninit( SSD1306DATA *p_data )
{
	if( p_data->fd == -1 ) {
		return ;
	}

	write(p_data->fd, "\x00\xae", 2) ;

	close( p_data->fd ) ;

	p_data->fd = -1 ;
}


// -------------------------------------------------------------------------------- 
//!  更新                                                                           
//!                                                                                 
//!  @param [in] p_data  SSD1306 データ                                             
//!  @retval      0      正常終了                                                   
//!  @retval     -1      異常終了                                                   
// -------------------------------------------------------------------------------- 
int ssd1306_update( SSD1306DATA *p_data )
{
	static const uint8_t	command[][4]={
								{ 0x00,0x22,0x00,0x07 },	// Set Page Address
								{ 0x00,0x21,0x00,0x7f },	// Set Column Address
							} ;
	int						result ;


	result = write(p_data->fd, &command[0][0], 4) ;
	if( result != 4 ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	result = write(p_data->fd, &command[1][0], 4) ;
	if( result != 4 ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	result = write(p_data->fd, &p_data->control, 1+sizeof( p_data->vram.b )) ;
	if( result != (1+sizeof( p_data->vram.b )) ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  表示制御                                                                       
//!                                                                                 
//!  @param [in] p_data  SSD1306 データ                                             
//!  @param [in] on       true:表示オン                                             
//!                      false:表示オフ                                             
//!  @retval      0      正常終了                                                   
//!  @retval     -1      異常終了                                                   
// -------------------------------------------------------------------------------- 
int ssd1306_display(SSD1306DATA *p_data, bool on)
{
	uint8_t		command[2] ;
	int			result ;


	if( p_data->fd == -1 ) {
		return -1 ;
	}

	command[0] = 0x00 ;
	command[1] = 0xae | (on != false) ;

	result = write(p_data->fd, &command[0], sizeof( command )) ;
	if( result != sizeof( command ) ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  コントラスト設定                                                               
//!                                                                                 
//!  @param [in] p_data  SSD1306 データ                                             
//!  @param [in] value   値                                                         
//!  @retval      0      正常終了                                                   
//!  @retval     -1      異常終了                                                   
// -------------------------------------------------------------------------------- 
int ssd1306_contrast(SSD1306DATA *p_data, uint8_t value)
{
	uint8_t		command[3] ;
	int			result ;


	command[0] = 0x00 ;
	command[1] = 0x81 ;
	command[2] = value ;
	result     = write(p_data->fd, &command[0], sizeof( command )) ;
	if( result != sizeof( command ) ) {
		fprintf(stderr, "******** %s(%d)\n", __func__, __LINE__) ;
		return -1 ;
	}

	return 0 ;
}


// -------------------------------------------------------------------------------- 
//!  VRAM クリア                                                                    
//!                                                                                 
//!  @param [out] p_data  SSD1306 データ                                            
//!  @return              なし                                                      
// -------------------------------------------------------------------------------- 
void ssd1306_clear( SSD1306DATA *p_data )
{
	memset(&p_data->vram.b[0][0], 0, sizeof( p_data->vram.b )) ;
}


// -------------------------------------------------------------------------------- 
//!  クリア（行単位）                                                               
//!                                                                                 
//!  @param [in,out] p_data  SSD1306 データ                                         
//!  @param [in]     line    開始行                                                 
//!  @param [in]     x       開始ｘ座標                                             
//!  @param [in]     width   描画幅                                                 
//!  @param [in]     height  描画高さ                                               
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
void ssd1306_clearLine(SSD1306DATA *p_data, int line, int x, int width, int height)
{
	uint8_t		*p_dest, *p_destEnd ;


	while( height-- ) {
		p_dest    = &p_data->vram.b[x][line] ;
		p_destEnd = &p_data->vram.b[x+width][line] ;

		do {
			p_dest[0] = 0 ;
			p_dest   += SSD1306_LINE ;
		} while( p_dest < p_destEnd ) ;

		line++ ;
	}
}


// -------------------------------------------------------------------------------- 
//!  画像を描画（行単位）                                                           
//!                                                                                 
//!  @param [in,out] p_data   SSD1306 データ                                        
//!  @param [in]     line     開始行                                                
//!  @param [in]     x        開始ｘ座標                                            
//!  @param [in]     width    画像幅                                                
//!  @param [in]     height   画像高さ                                              
//!  @param [in]     p_image  画像データ                                            
//!  @return                  なし                                                  
// -------------------------------------------------------------------------------- 
void ssd1306_drawImageLine(SSD1306DATA *p_data, int line, int x, int width, int height, const uint8_t *p_image)
{
	uint8_t		*p_dest, *p_destEnd ;


	while( height-- ) {
		p_dest    = &p_data->vram.b[x][line] ;
		p_destEnd = &p_data->vram.b[x+width][line] ;

		do {
			p_dest[0] = *p_image++ ;
			p_dest   += SSD1306_LINE ;
		} while( p_dest < p_destEnd ) ;

		line++ ;
	}
}


// -------------------------------------------------------------------------------- 
//!  画像を反転（行単位）                                                           
//!                                                                                 
//!  @param [in,out] p_data  SSD1306 データ                                         
//!  @param [in]     line    開始行                                                 
//!  @param [in]     x       開始ｘ座標                                             
//!  @param [in]     width   描画幅                                                 
//!  @param [in]     height  描画高さ                                               
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
void ssd1306_invertLine(SSD1306DATA *p_data, int line, int x, int width, int height)
{
	uint8_t		*p_dest, *p_destEnd ;


	while( height-- ) {
		p_dest    = &p_data->vram.b[x][line] ;
		p_destEnd = &p_data->vram.b[x+width][line] ;

		do {
			p_dest[0] ^= 0xff ;
			p_dest    += SSD1306_LINE ;
		} while( p_dest < p_destEnd ) ;

		line++ ;
	}
}


// -------------------------------------------------------------------------------- 
//!  画像をグレー（行単位）                                                         
//!                                                                                 
//!  @param [in,out] p_data  SSD1306 データ                                         
//!  @param [in]     line    開始行                                                 
//!  @param [in]     x       開始ｘ座標                                             
//!  @param [in]     width   描画幅                                                 
//!  @param [in]     height  描画高さ                                               
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
void ssd1306_grayLine(SSD1306DATA *p_data, int line, int x, int width, int height)
{
	uint8_t		mask, *p_dest, *p_destEnd ;


	mask = 0x55 ;
	if( x & 1 ) {
		mask = 0xaa ;
	}

	while( height-- ) {
		p_dest    = &p_data->vram.b[x][line] ;
		p_destEnd = &p_data->vram.b[x+width][line] ;

		do {
			p_dest[0] &= mask ;
			p_dest    += SSD1306_LINE ;
			mask       = (mask >> 1) | (mask << 7) ;
		} while( p_dest < p_destEnd ) ;

		line++ ;
	}
}


// -------------------------------------------------------------------------------- 
//!  線を描画                                                                       
//!                                                                                 
//!  @param [in,out] p_data  SSD1306 データ                                         
//!  @param [in]     sx      開始ｘ座標                                             
//!  @param [in]     sy      開始ｙ座標                                             
//!  @param [in]     ex      終了ｘ座標                                             
//!  @param [in]     ey      終了ｙ座標                                             
//!  @return                 なし                                                   
// -------------------------------------------------------------------------------- 
void ssd1306_line(SSD1306DATA *p_data, int sx, int sy, int ex, int ey)
{
	uint8_t		pattern, *p_dest ;
	int			temp, width, height, add, frac, count ;

	// 高さ
	height = ey - sy ;
	if( height < 0 ) {
		height = -height ;
		sy     = ey ;

		temp = sx ;
		sx   = ex ;
		ex   = temp ;
	}

	// 幅
	width = ex - sx ;
	add   = SSD1306_LINE ;
	if( width < 0 ) {
		width = -width ;
		add   = -SSD1306_LINE ;
	}

	p_dest = &p_data->vram.b[sx][sy/8] ;

	pattern = 1 << (sy & 7) ;

	if( width >= height ) {		// ｘ方向

		frac  = width ;
		count = width + 1 ;

		do {
			*p_dest |= pattern ;
			p_dest  += add ;

			frac -= height ;
			if( frac < width ) {
				frac   += width ;

				pattern <<= 1 ;
				if( !pattern ) {
					pattern = 1 ;
					p_dest++ ;
				}
			}
		} while( --count ) ;

	} else {					// ｙ方向

		frac  = height ;
		count = height + 1 ;

		do {
			*p_dest |= pattern ;

			pattern <<= 1 ;
			if( !pattern ) {
				pattern = 1 ;
				p_dest++ ;
			}

			frac -= width ;
			if( frac < height ) {
				frac   += height ;
				p_dest += add ;
			}
		} while( --count ) ;

	}
}
